"""
WiP.

Soon.
"""

# region [Imports]

import os
import re
import sys
import json
import queue
import math
import base64
import pickle
import random
import shelve
import dataclasses
import shutil
import asyncio
import logging
import sqlite3
import platform

import subprocess
import inspect

from time import sleep, process_time, process_time_ns, perf_counter, perf_counter_ns
from io import BytesIO, StringIO
from abc import ABC, ABCMeta, abstractmethod
from copy import copy, deepcopy
from enum import Enum, Flag, auto, unique
from pprint import pprint, pformat
from pathlib import Path
from string import Formatter, digits, printable, whitespace, punctuation, ascii_letters, ascii_lowercase, ascii_uppercase
from timeit import Timer
from typing import (TYPE_CHECKING, Mapping, TypeVar, TypeGuard, TypeAlias, Final, TypedDict, Generic, Union, Optional, ForwardRef, final, Callable,
                    no_type_check, no_type_check_decorator, overload, get_type_hints, cast, Protocol, runtime_checkable, NoReturn, NewType, Literal, AnyStr, IO, BinaryIO, TextIO, Any)
from collections import Counter, ChainMap, deque, namedtuple, defaultdict
from collections.abc import (AsyncGenerator, AsyncIterable, AsyncIterator, Awaitable, ByteString, Callable, Collection, Container, Coroutine, Generator,
                             Hashable, ItemsView, Iterable, Iterator, KeysView, Mapping, MappingView, MutableMapping, MutableSequence, MutableSet, Reversible, Sequence, Set, Sized, ValuesView)
from zipfile import ZipFile, ZIP_LZMA
from datetime import datetime, timezone, timedelta
from tempfile import TemporaryDirectory
from textwrap import TextWrapper, fill, wrap, dedent, indent, shorten
from functools import wraps, partial, lru_cache, singledispatch, total_ordering, cached_property, cache
from contextlib import contextmanager, asynccontextmanager, nullcontext, closing, ExitStack, suppress
from statistics import mean, mode, stdev, median, variance, pvariance, harmonic_mean, median_grouped
from urllib.parse import urlparse
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, wait, as_completed, ALL_COMPLETED, FIRST_EXCEPTION, FIRST_COMPLETED

from rich.console import Console as RichConsole
from rich.theme import Theme
from rich.style import Style, StyleType
from rich.table import Table

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


# from typing import is_typeddict
from inspect import iscoroutine, isasyncgen, isawaitable, isc
from asyncio import iscoroutine, isawaitable
if TYPE_CHECKING:
    ...

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class GidTheme(Theme):

    def __init__(self,
                 name: str = None,
                 extra_styles: Mapping[str, StyleType] | None = None):

        self._name = name or self.__class__.__name__
        self._extra_styles = extra_styles or {}
        super().__init__(self._generate_styles(), inherit=True)

    @property
    def name(self) -> str:
        return self._name

    def _generate_styles(self) -> dict[str, StyleType]:
        _out = {}

        return _out | self._extra_styles

    @property
    def config_table(self) -> Table:
        table = Table("Style-name", "Style-value",
                      title=self.name,
                      row_styles=["on grey11", "on grey15"])

        for style_name, style_value in sorted(self.styles.items()):
            table.add_row(style_name, f"[{style_value}]'{style_value!s}'[/{style_value}]")

        return table

# region [Main_Exec]


if __name__ == '__main__':
    x = RichConsole(soft_wrap=True, theme=GidTheme())
    from rich.text import Text
    from rich.syntax import Syntax, PygmentsSyntaxTheme
    from rich.panel import Panel

    with THIS_FILE_DIR.joinpath("rich_helper.py").open("r", encoding='utf-8', errors='ignore') as f:
        t = Syntax(f.read(), "python", theme=PygmentsSyntaxTheme("dracula"), line_numbers=True, background_color="black", indent_guides=True)
    p = Panel(t)
    x.print(p)
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import QIcon, QPixmap

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gidapptools_qt._data.images import PLACEHOLDER_IMAGE, DEFAULT_APP_ICON_IMAGE, StoredImage
from gidapptools.gidapptools_qt.resources.resources_helper import PixmapResourceItem

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class GidQtPlaceholderImage:

    def __init__(self, stored_image: StoredImage) -> None:
        self.stored_image = stored_image
        self._pixmap: QPixmap = None
        self._icon: QIcon = None

    @property
    def pixmap(self) -> QPixmap:
        if self._pixmap is None:
            self._pixmap = QPixmap(self.stored_image.path)
        return self._pixmap

    @property
    def icon(self) -> QIcon:
        if self._icon is None:
            self._icon = QIcon(self.pixmap)
        return self._icon


QT_PLACEHOLDER_IMAGE = PixmapResourceItem(PLACEHOLDER_IMAGE.path)
QT_DEFAULT_APP_ICON_IMAGE = PixmapResourceItem(DEFAULT_APP_ICON_IMAGE.path)
# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING, Callable
from hashlib import sha1, blake2b
from pathlib import Path

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.custom_types import PATH_TYPE
    from hashlib import _Hash

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


# FILE_HASH_INCREMENTAL_THRESHOLD: int = 104857600  # 100mb
FILE_HASH_INCREMENTAL_THRESHOLD: int = 52428800  # 50mb


def _actual_hash_file(in_file: Path, hash_algo: Callable = blake2b) -> "_Hash":
    if in_file.stat().st_size > FILE_HASH_INCREMENTAL_THRESHOLD:
        _hash = hash_algo(usedforsecurity=False)
        with in_file.open("rb", buffering=FILE_HASH_INCREMENTAL_THRESHOLD // 4) as f:
            for chunk in f:
                _hash.update(chunk)
        return _hash

    return hash_algo(in_file.read_bytes(), usedforsecurity=False)


def file_hash(in_file: "PATH_TYPE", hash_algo: Callable = blake2b) -> str:
    in_file = Path(in_file)
    if not in_file.is_file():
        raise OSError(f"The path {in_file.as_posix()!r} either does not exist or is a Folder.")
    return _actual_hash_file(in_file=in_file, hash_algo=hash_algo).hexdigest()


def hash_to_int(in_hash: str) -> int:
    return int(in_hash, base=16)

# region [Main_Exec]


if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from abc import ABC, abstractmethod
from typing import Any
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.meta_data.config_kwargs import ConfigKwargs
from gidapptools.abstract_classes.abstract_meta_item import AbstractMetaItem

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class AbstractMetaFactory(ABC):
    product_class: AbstractMetaItem = None
    default_configuration: dict[str, Any] = {}

    def __init__(self, config_kwargs: ConfigKwargs) -> None:
        if self.product_class is None:
            raise TypeError("Can't instantiate abstract class MetaInfoFactory with abstract class attribute 'product_class'")

        self.is_setup = False
        self.config_kwargs = config_kwargs

    @classmethod
    @property
    def __default_configuration__(cls) -> dict[str, Any]:

        return cls.default_configuration | cls.product_class.__default_configuration__

    @classmethod
    @property
    def product_name(cls) -> str:
        return cls.product_class.name

    @abstractmethod
    def setup(self) -> None:
        self.is_setup = True

    @abstractmethod
    def _build(self) -> AbstractMetaItem:
        ...

    @classmethod
    def build(cls, config_kwargs: ConfigKwargs) -> AbstractMetaItem:
        factory_instance = cls(config_kwargs=config_kwargs)
        instance = factory_instance._build()
        instance.to_storager(factory_instance.config_kwargs.get('storager'))
        factory_instance.config_kwargs.created_meta_items[factory_instance.product_name] = instance
        return instance


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING
from pathlib import Path
from zipfile import ZIP_LZMA, ZipFile
from multiprocessing import Process

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.custom_types import PATH_TYPE

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def compress_file(source: "PATH_TYPE", target: "PATH_TYPE", suffix: str = '.zip'):
    source = Path(source)

    target = Path(target)
    zip_target = target.with_suffix(suffix)
    with ZipFile(zip_target, "w", ZIP_LZMA) as zippy:
        zippy.write(source, source.name)


def compress_in_process(source: "PATH_TYPE", target: "PATH_TYPE", suffix: str = '.zip') -> Process:
    process = Process(daemon=False, target=compress_file, kwargs={"source": source, "target": target, "suffix": suffix})
    process.start()
    return process


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
from typing import TYPE_CHECKING, Any, Union, Literal, Hashable, Iterable, Optional
from pathlib import Path
from datetime import timezone, datetime
from contextlib import contextmanager

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.general_helper.date_time import DateTimeFrame
    from gidapptools.gid_config.parser.config_data import ConfigData
    from gidapptools.gid_config.conversion.spec_item import SpecEntry
    from gidapptools.meta_data.meta_paths.meta_paths_item import NamedMetaPath

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def show_basic_error_pop_up(message: str, title: str = None) -> None:
    from tkinter import messagebox
    messagebox.showerror(title=title or "Error", message=message)


class GidAppToolsBaseError(Exception):
    """
    Base Exception For GidAppTools.
    """


class GidAppToolsFatalError(RuntimeError):
    ...


class ConversionError(GidAppToolsBaseError):
    ...


class NotUsableError(GidAppToolsBaseError):

    def __init__(self,
                 subject: object | type,
                 reason: Literal["unfinished", "broken", "unknown"] | None = None) -> None:
        self.subject = subject
        self.reason = reason if reason is not None else "unknown"
        self.msg = f"{self.subject!r} is not usable, reason: {self.reason!r}."
        super().__init__(self.msg)


class ApplicationInstanceAlreadyRunningError(GidAppToolsFatalError):
    def __init__(self, app_name: str, running_pid: int) -> None:
        self.app_name = app_name
        self.running_pid = running_pid
        os.environ["FATAL_ERROR_RAISED"] = "1"
        super().__init__(f"There is already an instance of {self.app_name!r} running with the pid of {self.running_pid!r}.")


class MissingOptionalDependencyError(GidAppToolsBaseError):

    def __init__(self, dependency_name: str, package_name: str = None) -> None:
        self.dependency_name = dependency_name
        self.package_name = package_name
        self.msg = f"Missing optional dependency {self.dependency_name!r}"
        if self.package_name is not None:
            self.msg += f" try installing '{self.package_name!s}[{self.dependency_name!s}]'"
        self.msg += "."
        super().__init__(self.msg)

    @classmethod
    @contextmanager
    def try_import(cls, package_name: str = None):
        try:
            yield
        except ImportError as e:
            if e.path is None:
                dependency_name = e.name
                raise cls(dependency_name=dependency_name, package_name=package_name) from e

            raise e


class FlagConflictError(GidAppToolsBaseError):
    def __init__(self, flag_names: Iterable[str], conflicting_bool_value: bool = True):
        self.flag_names = tuple(flag_names)
        self.conflicting_bool_value = conflicting_bool_value
        self.msg = self._build_msg()

        super().__init__(self.msg)

    def _build_msg(self) -> str:
        text = "No more than one of "
        flags = list(self.flag_names)
        last_flag = flags.pop(-1)
        text += ', '.join(flags) + ' or ' + repr(last_flag)
        text += f" can be {self.conflicting_bool_value!r}."
        return text


class DispatchError(GidAppToolsBaseError):
    ...

# TODO: accept AdvancedDict instance as parameter and show path_taken maybe.


class DictMergeConflict(GidAppToolsBaseError):
    def __init__(self, first_dict: dict, second_dict: dict, conflicting_key: Hashable, none_values: set[Hashable]) -> None:
        self.first_dict = first_dict
        self.second_dict = second_dict
        self.conflicting_key = conflicting_key
        self.none_values = none_values
        self.message = f"Unable to merge key {self.conflicting_key!r}, because it already exists in the first dictionary and has an value that is not a 'none_value'({', '.join(f'{item!r}' for item in self.none_values)}"
        super().__init__(self.message)


class AdvancedDictError(GidAppToolsBaseError):
    ...


class GidConfigError(GidAppToolsBaseError):
    ...


class GidQtError(GidAppToolsBaseError):
    ...


class ApplicationNotExistingError(GidQtError):

    def __init__(self, msg: str = None) -> None:
        self.msg = msg or "No Application instance found."
        super().__init__(self.msg)


class ApplicationExistsError(GidQtError):

    def __init__(self, existing_application, msg: str) -> None:
        self.existing_application = existing_application
        self.msg = f"{msg} -> application: {self.existing_application!r}"
        super().__init__(self.msg)


class ApplicationNotSetupError(GidQtError):
    ...


class SectionExistsError(GidConfigError):
    ...


class SectionMissingError(GidConfigError):

    def __init__(self, section_name: str, config_data: "ConfigData") -> None:
        self.section_name = section_name
        self.config_data = config_data
        self.message = f"No Section named {self.section_name!r} in config_data {self.config_data}."
        super().__init__(self.message)


class EntryMissingError(GidConfigError):
    def __init__(self, section_name: str, entry_key: str, config_data: "ConfigData") -> None:
        self.section_name = section_name
        self.entry_key = entry_key
        self.config_data = config_data
        self.message = f"No Entry with key {self.entry_key!r} in Section named {self.section_name!r} in config_data {self.config_data}."
        super().__init__(self.message)


class ValueValidationError(GidConfigError):

    def __init__(self, config_value: Any, spec_entry: "SpecEntry", validation_description: str = None) -> None:
        self.config_value = config_value
        self.spec_entry = spec_entry
        self.validation_description = validation_description
        self.msg = f"Value {self.config_value!r} with Spec {self.spec_entry!r} failed its validation"

        if self.validation_description is not None:
            self.msg += f", {self.validation_description}"
        self.msg += "."

        super().__init__(self.msg)


class ConfigSpecError(GidConfigError):
    ...


class InvalidConverterValue(ConfigSpecError):
    ...


class UnconvertableTypusError(ConfigSpecError):
    ...


class SpecDataMissingError(ConfigSpecError):
    ...


class ConversionError(GidConfigError):
    ...


class MissingTypusOrSpecError(GidConfigError):
    ...


class IniParsingError(GidConfigError):
    ...


class MissingDefaultValue(GidConfigError):
    ...


class TrailingCommentError(IniParsingError):
    ...


class NotLoggingLevelError(GidConfigError):

    def __init__(self, value, converter, section_name: str = None, entry_name: str = None, config=None) -> None:
        self.value = value

        self.converter = converter
        self.section_name = section_name
        self.entry_name = entry_name
        self.config = config
        self.msg = f"Entry {self.entry_name!r} of section {self.section_name}, of config {self.config!r} has a value ({self.value!r}) that is not a valid value of converter {self.converter!r}."
        super().__init__(self.msg)


class MinError(GidConfigError):

    def __init__(self, value, min_value, converter, section_name: str = None, entry_name: str = None, config=None) -> None:
        self.value = value
        self.min_value = min_value
        self.converter = converter
        self.section_name = section_name
        self.entry_name = entry_name
        self.config = config
        self.msg = f"Entry {self.entry_name!r} of section {self.section_name}, of config {self.config!r} has a value ({self.value!r}) that is less than the min_value({self.min_value!r}) of converter {self.converter!r}."
        super().__init__(self.msg)


class MaxError(GidConfigError):

    def __init__(self, value, max_value, converter, section_name: str = None, entry_name: str = None, config=None) -> None:
        self.value = value
        self.max_value = max_value
        self.converter = converter
        self.section_name = section_name
        self.entry_name = entry_name
        self.config = config
        self.msg = f"Entry {self.entry_name!r} of section {self.section_name}, of config {self.config!r} has a value ({self.value!r}) that is less than the max_value({self.max_value!r}) of converter {self.converter!r}."
        super().__init__(self.msg)


class KeyPathError(AdvancedDictError):
    def __init__(self, missing_key: Hashable, key_path: list[Hashable], last_key: Hashable = None) -> None:
        self.missing_key = missing_key
        self.last_key = last_key
        self.key_path = key_path if self.last_key is None else key_path + [last_key]
        self.missing_key_pos = self.key_path.index(self.missing_key) + 1
        self.missing_key_pos_verbose = f"{self.missing_key_pos}."
        self.message = f"The {self.missing_key_pos_verbose} key {self.missing_key!r} was not found, full key_path: {self.key_path!r}."
        super().__init__(self.message)


class NotMappingError(AdvancedDictError):
    def __init__(self, key: Hashable, value: Any, key_path: list[Hashable], action: Union[Literal['set'], Literal['get']], last_key: Hashable = None) -> None:
        self.key = key
        self.value = value
        self.value_type = type(value)
        self.action = action
        self.last_key = last_key
        self.key_path = key_path if self.last_key is None else key_path + [last_key]
        self.key_pos = self.key_path.index(self.key) + 1
        self.key_pos_verbose = f"{self.key_pos}."
        self.message = f"Unable to {self.action} {self.key_pos_verbose} {self.key!r} (key_path={self.key_path!r}), because value is not a Mapping but {self.value_type!r}."
        super().__init__(self.message)


class DateTimeFrameTimezoneError(GidAppToolsBaseError):
    def __init__(self, duration_item: "DateTimeFrame", start_tz: Optional[timezone], end_tz: Optional[timezone], message: str) -> None:
        self.duration_item = duration_item
        self.start_tz = start_tz
        self.end_tz = end_tz
        self.message = message + f", {start_tz=}, {end_tz=}."
        super().__init__(self.message)


class NotUtcDatetimeError(GidAppToolsBaseError):
    def __init__(self, datetime_obj: datetime) -> None:
        self.datetime_obj = datetime_obj
        self.actual_timezone = datetime_obj.tzinfo
        self.message = f"{self.datetime_obj!r} has to be of timezone {timezone.utc!r}, not {self.actual_timezone!r}"
        super().__init__(self.message)


class BaseMetaDataError(GidAppToolsBaseError):
    """
    Base Error for Meta Data Errors.
    """


class RegisterAfterSetupError(BaseMetaDataError):
    ...


class MetaItemNotFoundError(BaseMetaDataError):
    def __init__(self, requested_name: str, existing_item_names: list[str]) -> None:
        self.requested_name = requested_name
        self.existing_item_names = existing_item_names
        self.message = f"Item {self.requested_name!r} was not found in existing meta_items, existing_meta_item names: {self.existing_item_names!r}"
        super().__init__(self.message)


class NoFactoryFoundError(BaseMetaDataError):
    def __init__(self, app_meta_item_name) -> None:
        self.app_meta_item_name = app_meta_item_name
        self.message = f"Unable to find a factory for {self.app_meta_item_name!r}."
        super().__init__(self.message)


class NotSetupError(BaseMetaDataError):
    def __init__(self, app_meta_data) -> None:
        self.app_meta_data = app_meta_data
        self.message = f'{self.app_meta_data.__class__.__name__!r} has to set up from the base_init_file first!'
        super().__init__(self.message)


class NotBaseInitFileError(BaseMetaDataError):
    """
    Raised if the calling file is not the base __init__.py file of the App.
    """

    def __init__(self, calling_file: Path) -> None:
        self.calling_file = calling_file
        self.calling_folder = calling_file.parent
        self.message = f"This function has to be called from the '__init__.py' file in the base directory of the App, actual calling file: {self.calling_file.as_posix()!r}"
        super().__init__(self.message)


class BaseMetaPathsError(GidAppToolsBaseError):
    """
    Base error for meta_paths.
    """


class UnknownMetaPathIdentifier(BaseMetaPathsError):

    def __init__(self, identifier: str) -> None:
        self.identifier = identifier
        self.msg = f"{self.identifier!r} is not an valid identifier for a Meta-Path."
        super().__init__(self.msg)


class NotImplementedMetaPath(BaseMetaPathsError):

    def __init__(self, identifier: "NamedMetaPath") -> None:
        self.identifier = identifier
        self.msg = f"No Path implemented for identifer {self.identifier!r}."
        super().__init__(self.msg)


class AppNameMissingError(BaseMetaPathsError):

    def __init__(self) -> None:
        self.message = "The name of the App was not found in the 'kwargs_holder' or in the env ('APP_NAME')."
        super().__init__(self.message)


class UnparsableHumanTimedelta(ConversionError):

    def __init__(self, human_text: str) -> None:
        self.human_text = human_text
        self.message = f"Unable to parse the text {self.human_text!r} to a valid timedelta."
        super().__init__(self.message)


# region [Main_Exec]
if __name__ == '__main__':
    pass


# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import sys
from typing import TYPE_CHECKING, Union, Literal, TypeVar, Optional, TypeAlias

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

# * Standard Library Imports ---------------------------------------------------------------------------->
from enum import Enum, auto, unique
from pathlib import Path
from functools import partial
from threading import RLock
from collections.abc import Mapping, Callable

# * Third Party Imports --------------------------------------------------------------------------------->

import apsw
import peewee
from frozendict import frozendict
from playhouse.apsw_ext import APSWDatabase

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_database.orm_peewee.sqlite.constants import MEMORY_DB_PATH, STD_DEFAULT_PRAGMAS, STD_DEFAULT_EXTENSIONS
from gidapptools.gid_database.orm_peewee.sqlite.pragma_info import PragmaInfo
from gidapptools.gid_logger.logger import get_logger

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.custom_types import PATH_TYPE

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()
log = get_logger(__name__)

# endregion [Constants]

# region [Types]

WAL_HOOK_TYPE: TypeAlias = Optional[Callable[["GidAPSWDatabase", apsw.Connection, str, int], int]]

PROFILE_HOOK_TYPE: TypeAlias = Optional[Callable[["GidAPSWDatabase", str, int], None]]

UPDATE_HOOK_TYPE: TypeAlias = Optional[Callable[["GidAPSWDatabase", int, str, str, int], None]]

ROLLBACK_HOOK_TYPE: TypeAlias = Optional[Callable[["GidAPSWDatabase"], None]]

COMMIT_HOOK_TYPE: TypeAlias = Optional[Callable[["GidAPSWDatabase"], None]]

ALL_HOOK_TYPES: TypeAlias = Union[WAL_HOOK_TYPE, PROFILE_HOOK_TYPE, UPDATE_HOOK_TYPE, ROLLBACK_HOOK_TYPE, COMMIT_HOOK_TYPE]

# endregion [Types]


def _ensure_string_db_path(in_db_path: Union["PATH_TYPE", Literal[":memory:"]]) -> str:
    """
    Returns the provided path resolved and as a String so APSW can understand it.

    Peewee ( or APSW) seems to not follow the `os.PathLike`- Protocol.

    :param in_db_path: _description_
    :type in_db_path: Union["PATH_TYPE", Literal[":memory:"]]
    :return: _description_
    :rtype: str
    """
    if in_db_path == MEMORY_DB_PATH:
        return MEMORY_DB_PATH
    db_path = Path(in_db_path).resolve()
    return os.fspath(db_path)


def data_setup_noop(db: "GidAPSWDatabase") -> None:
    ...


def migration_noop(db: "GidAPSWDatabase") -> None:
    ...


class DatabaseSetupStatus(Enum):
    NOT_SETUP = auto()
    SETUP_RUNNING = auto()
    OPERATIONAL = auto()
    SHUTDOWN = auto()


@unique
class APSWHooks(Enum):
    WAL_HOOK = ("_wal_hook", "setwalhook")
    PROFILE_HOOK = ("_profile_hook", "setprofile")
    UPDATE_HOOK = ("_update_hook", "setupdatehook")
    ROLLBACK_HOOK = ("_rollback_hook", "setrollbackhook")
    COMMIT_HOOK = ("_commit_hook", "setcommithook")

    @property
    def attribute_name(self) -> str:
        return self.value[0]

    @property
    def setter_name(self) -> str:
        return self.value[1]

    @classmethod
    def _missing_(cls, value: object) -> "APSWHooks":
        if isinstance(value, str):
            for name, item in cls.__members__.items():

                if name == value.upper() or value.casefold() in item.value or value.casefold() in [i.removeprefix("_") for i in item.value]:
                    return item

        return super()._missing_(value)


WRAP_WITH_DB_TYPEVAR = TypeVar("WRAP_WITH_DB_TYPEVAR", WAL_HOOK_TYPE, PROFILE_HOOK_TYPE, UPDATE_HOOK_TYPE, ROLLBACK_HOOK_TYPE, COMMIT_HOOK_TYPE, None)


class APSWHookManager:
    __slots__ = ("database", "_wal_hook", "_profile_hook", "_update_hook", "_rollback_hook", "_commit_hook")

    def __init__(self,
                 database: "GidAPSWDatabase",
                 wal_hook: WAL_HOOK_TYPE = None,
                 profile_hook: PROFILE_HOOK_TYPE = None,
                 update_hook: UPDATE_HOOK_TYPE = None,
                 rollback_hook: ROLLBACK_HOOK_TYPE = None,
                 commit_hook: COMMIT_HOOK_TYPE = None) -> None:
        self.database = database
        self._wal_hook = wal_hook
        self._profile_hook = profile_hook
        self._update_hook = update_hook
        self._rollback_hook = rollback_hook
        self._commit_hook = commit_hook

    def _wrap_with_db(self, orig_func: WRAP_WITH_DB_TYPEVAR) -> WRAP_WITH_DB_TYPEVAR:
        if orig_func is not None:
            return partial(orig_func, self.database)

        return orig_func

    def apply(self, connection: apsw.Connection) -> None:
        for hook_info in APSWHooks.__members__.values():
            getattr(connection, hook_info.setter_name)(self._wrap_with_db(getattr(self, hook_info.attribute_name, None)))

    def set(self, typus: Union[str, APSWHooks], value: WRAP_WITH_DB_TYPEVAR) -> WRAP_WITH_DB_TYPEVAR:
        typus = APSWHooks(typus)
        setattr(self, typus.attribute_name, value)
        if not self.database.is_closed():
            getattr(self.database.connection(), typus.setter_name)(self._wrap_with_db(getattr(self, typus.attribute_name, None)))
        return value

    def set_wal_hook(self, value: WAL_HOOK_TYPE) -> WAL_HOOK_TYPE:
        self.set(typus=APSWHooks.WAL_HOOK, value=value)
        return value

    def set_profile_hook(self, value: PROFILE_HOOK_TYPE) -> PROFILE_HOOK_TYPE:
        self.set(typus=APSWHooks.PROFILE_HOOK, value=value)
        return value

    def set_update_hook(self, value: UPDATE_HOOK_TYPE) -> UPDATE_HOOK_TYPE:
        self.set(typus=APSWHooks.UPDATE_HOOK, value=value)
        return value

    def set_rollback_hook(self, value: ROLLBACK_HOOK_TYPE) -> ROLLBACK_HOOK_TYPE:
        self.set(typus=APSWHooks.ROLLBACK_HOOK, value=value)
        return value

    def set_commit_hook(self, value: COMMIT_HOOK_TYPE) -> COMMIT_HOOK_TYPE:
        self.set(typus=APSWHooks.COMMIT_HOOK, value=value)
        return value


class GidAPSWDatabase(APSWDatabase):
    setup_lock = RLock()
    default_pragmas: Mapping[str, object] = frozendict(**STD_DEFAULT_PRAGMAS)
    default_extensions: Mapping[str, bool] = frozendict(**STD_DEFAULT_EXTENSIONS)

    def __init__(self,
                 database_path: Union["PATH_TYPE", Literal[":memory:"]],
                 backup_folder: "PATH_TYPE" = None,
                 thread_safe: bool = True,
                 autoconnect: bool = True,
                 autorollback: bool = None,
                 timeout: Union[Callable[[int], bool], float, None] = 100,
                 pragmas: Mapping = None,
                 extensions: Mapping = None,
                 hooks: dict[str, ALL_HOOK_TYPES] = None):
        self.setup_status: DatabaseSetupStatus = DatabaseSetupStatus.NOT_SETUP
        self._backup_folder = Path(backup_folder).resolve() if backup_folder is not None else None
        self.hook_manager: APSWHookManager = APSWHookManager(self, **(hooks or {}))
        self.pragma_info: PragmaInfo = None
        super().__init__(database=_ensure_string_db_path(database_path),
                         autoconnect=autoconnect,
                         autorollback=autorollback,
                         thread_safe=thread_safe,
                         timeout=timeout,
                         pragmas=dict(self.default_pragmas | (pragmas or {})),
                         **dict(self.default_extensions | (extensions or {})))

    def init(self, database, pragmas=None, timeout=5, returning_clause=None, **kwargs):

        super().init(database, pragmas, timeout, returning_clause, **kwargs)
        self.timeout = timeout

    @property
    def db_file_name(self) -> str:
        if self.database == MEMORY_DB_PATH:
            return MEMORY_DB_PATH
        return self.db_path.name

    @property
    def db_path(self) -> Optional[Path]:
        if self.database == MEMORY_DB_PATH:
            return None
        return Path(self.database)

    @property
    def backup_folder(self) -> Optional[Path]:
        if self._backup_folder is None:
            return None

        return self._backup_folder

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value: Union[float, Callable[[int], bool], None]):
        if value is None:
            if self._timeout is value:
                return
            self._timeout = value

        elif callable(value):
            if self._timeout is value:
                return
            self._timeout = value
            if not self.is_closed():
                self.connection().setbusyhandler(self._timeout)

        else:
            if self._timeout == (value * 1000):
                return
            self._timeout = value * 1000

            if not self.is_closed():
                self.connection().setbusytimeout(self._timeout)

    def _connect(self):

        conn = apsw.Connection(self.database, **self.connect_params)
        if callable(self._timeout):
            conn.setbusyhandler(self._timeout)
        elif self._timeout is not None:
            conn.setbusytimeout(self._timeout * 1000)
        try:
            self._add_conn_hooks(conn)
        except:
            conn.close()
            raise
        return conn

    def _add_conn_hooks(self, conn: apsw.Connection):
        super()._add_conn_hooks(conn)
        self.hook_manager.apply(conn)
        wal_autocheckpoint = next((i[1] for i in self._pragmas if i[0] == "wal_autocheckpoint"), None)
        if wal_autocheckpoint is not None:
            conn.wal_autocheckpoint(wal_autocheckpoint)

    def vacuum(self, incremental: bool = True) -> bool:
        if self.is_closed() is True:
            return False

        try:
            conn: apsw.Connection = self.connection()
            if incremental is True:
                conn.execute("PRAGMA incremental_vacuum").fetchall()
                return True

            else:
                conn.execute("VACUUM").fetchall()
                return True
        except apsw.Error as e:
            return False

    def optimize(self, analysis_limit: int = 1000) -> bool:
        if self.is_closed() is True:
            return False

        try:
            conn: apsw.Connection = self.connection()
            conn.execute(f"PRAGMA analysis_limit={analysis_limit};PRAGMA optimize").fetchall()
            return True
        except apsw.Error as e:
            return False

    def _close(self, conn: apsw.Connection) -> None:
        if self.is_closed() is False:
            conn.execute("PRAGMA analysis_limit=0;PRAGMA optimize").fetchall()

        conn.close()

    def _setup_db_path(self, overwrite: bool = False) -> None:
        if self.db_path is not None:
            self.db_path.parent.mkdir(exist_ok=True, parents=True)
            if overwrite is True:
                self.db_path.unlink(missing_ok=True)

    def _set_page_size(self):
        log.debug("setting page size")
        wal_mode = False
        conn: apsw.Connection = self.connection()
        journal_mode = conn.execute("PRAGMA journal_mode;").fetchone()[0]
        if journal_mode.casefold() == "wal":
            wal_mode = True
        page_size = next((i[1] for i in self._pragmas if i[0] == "page_size"), None)

        if page_size is None:
            return

        conn.execute(f"PRAGMA page_size={page_size};").fetchall()

        if wal_mode is True:
            conn.execute("PRAGMA journal_mode=OFF;").fetchall()

        conn.execute("VACUUM;").fetchall()

        if wal_mode is True:
            conn.execute("PRAGMA journal_mode=WAL;").fetchall()

    def setup(self,
              db_proxy: peewee.DatabaseProxy = None,
              data_setup: Callable[["GidAPSWDatabase"], None] = data_setup_noop,
              migration: Callable[["GidAPSWDatabase"], None] = migration_noop,
              overwrite: bool = False) -> Self:
        with self.setup_lock:
            if self.setup_status is not DatabaseSetupStatus.NOT_SETUP:
                return self

            self.setup_status = DatabaseSetupStatus.SETUP_RUNNING
            self._pre_setup()
            self._setup_db_path(overwrite=overwrite)
            db_exists = self.db_path is not None and self.db_path.exists()
            self.connect()
            if db_exists is False:
                self._set_page_size()
            if db_proxy:
                db_proxy.initialize(self)

            data_setup(self)

            migration(self)

            self._post_setup()
            self.setup_status = DatabaseSetupStatus.OPERATIONAL
            return self

    def _pre_setup(self) -> None:
        ...

    def _post_setup(self) -> None:
        self.pragma_info = PragmaInfo(self).fill_with_data()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(database_name={self.db_file_name!r}, setup_status={self.setup_status.name!r})"


# region [Main_Exec]
if __name__ == '__main__':

    def run_setup(db):
        script_path = Path(r"D:\Dropbox\hobby\Modding\Programs\Github\My_Repos\GidAppTools\tests\gid_database\sqlite\data\simple_db_setup.sql")
        script = script_path.read_text(encoding='utf-8', errors='ignore')
        db.execute(peewee.SQL(script))
    x = GidAPSWDatabase(MEMORY_DB_PATH)
    x.setup(data_setup=run_setup)
    with x:
        print(x.execute(peewee.SQL('SELECT "Person"."name", "Country"."name" FROM "Person"  INNER JOIN "Country"  ON "Country"."id"=="Person"."country"')).fetchall())


# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import inspect
from functools import wraps
from typing import TYPE_CHECKING, Any, Union, Callable, Iterable
from pathlib import Path
from weakref import WeakSet, WeakMethod, ref
from gidapptools.errors import NotUsableError
# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from weakref import ReferenceType

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def ref_or_weakmethod(item, callback: Callable[["ReferenceType"], Any]) -> "ReferenceType":
    if inspect.ismethod(item):
        return WeakMethod(item, callback)
    return ref(item, callback)


class MethodEnabledWeakSet(WeakSet):

    def add(self, item) -> None:
        if self._pending_removals:
            self._commit_removals()
        ref_item = ref_or_weakmethod(item, self._remove)
        self.data.add(ref_item)


def make_repr(instance: object, attr_names: Union[Callable, Iterable[str]] = None, exclude_none: bool = True) -> str:
    attr_names = attr_names or vars
    if callable(attr_names):
        attr_dict = attr_names(instance)
    else:
        attr_dict = {}
        for name in attr_names:
            try:
                attr_value = getattr(instance, name)
                if callable(attr_value):
                    attr_value = attr_value()
            except Exception as e:

                attr_value = f"<{e.__class__.__name__}>"
            attr_dict[name] = attr_value

    if exclude_none is True:
        attr_dict = {k: v for k, v in attr_dict.items() if v is not None}

    return f"{instance.__class__.__name__}(" + ', '.join(f"{k}={v!r}" for k, v in attr_dict.items()) + ')'

# TODO: Unfinished


def UNFINISHED(_obj: Any):

    if hasattr(_obj, "__init__"):
        @wraps(_obj.__init__)
        def _inner(*args, **kwargs):

            raise NotUsableError(_obj, "unfinished")

        _obj.__init__ = _inner

        return _obj

    else:
        @wraps(_obj)
        def _inner(*args, **kwargs):

            raise NotUsableError(_obj, "unfinished")

        return _inner


@UNFINISHED
class ClassProperty:
    """
    # UNFINISHED

    rewrite probably needed
    """

    # TODO: researching how others have solved this!

    def __init__(self, fget=None, doc=None):
        self.fget = fget
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc
        self._name = ""

    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError(f'unreadable attribute {self._name}')
        return self.fget(obj)


# region [Main_Exec]
if __name__ == '__main__':
    print(f"{ClassProperty()=}")

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from enum import Flag, auto
from typing import Union
from pathlib import Path
from threading import RLock

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import QIcon, QImage, QMovie, QPixmap
from PySide6.QtCore import Qt, QFile, QSize, QByteArray

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def make_ressource_name(in_item_name: str, in_section_name: str = None) -> str:
    item_name = in_item_name.rsplit('.', 1)[0]
    if in_section_name:
        return f":/{in_section_name}/{item_name}"
    return f":/{item_name}"


class ResourceTypus(Flag):
    MISC = auto()
    PIXMAP = auto()
    MOVIE = auto()

    @classmethod
    def from_file_path(cls, file_path: Path) -> "ResourceTypus":
        suffix = file_path.suffix.casefold().strip('.')
        if suffix in {"bmp", "jpg", "jpeg", "png", "pbm", "pgm", "ppm", "xbm", "xpm", "svg"}:
            return cls.PIXMAP
        if suffix in {"gif"}:
            return cls.MOVIE
        return cls.MISC


class ResourceItem:
    cache_lock = RLock()
    _cache = {}

    def __init__(self, file_path: str = None, qt_path: str = None) -> None:
        self.file_path = Path(file_path) if file_path is not None else None
        self.qt_path = qt_path
        self.prefixes, self.name = self._get_qt_path_parts()

    def get_from_cache(self, key):
        with self.cache_lock:
            return self._cache.get(key, None)

    def store_in_cache(self, key, value):
        with self.cache_lock:
            self._cache[key] = value

    def _get_qt_path_parts(self) -> str:
        if self.qt_path is not None:
            as_path = Path(self.qt_path)
            _, *prefixes, name = as_path.parts
            return tuple(prefixes), name.rsplit('.')[0]
        else:
            name = self.file_path.stem
            prefixes = tuple()
            return prefixes, name

    def get_as_file(self) -> QFile:
        from_cache = self.get_from_cache(("file", self.name))
        if from_cache is not None:
            return from_cache

        _out = QFile(self.qt_path or str(self.file_path))
        self.store_in_cache(("image", self.name), _out)
        return _out

    @classmethod
    def clear_cache(cls):
        with cls.cache_lock:
            cls._cache.clear()


class MiscResourceItem(ResourceItem):
    ...


class PixmapResourceItem(ResourceItem):

    def get_as_pixmap(self, width=None, height=None) -> QPixmap:
        from_cache = self.get_from_cache(("pixmap", self.name, width, height))
        if from_cache is not None:
            return from_cache

        pixmap = QPixmap(self.qt_path or str(self.file_path))
        if any([width is None, height is None]):
            _out = pixmap
        else:
            _out = pixmap.scaled(QSize(width, height), Qt.KeepAspectRatioByExpanding, Qt.SmoothTransformation)
        self.store_in_cache(("pixmap", self.name, width, height), _out)
        return _out

    def get_as_icon(self) -> QIcon:
        from_cache = self.get_from_cache(("icon", self.name))
        if from_cache is not None:
            return from_cache
        _out = QIcon(self.qt_path or str(self.file_path))
        self.store_in_cache(("icon", self.name), _out)
        return _out

    def get_as_image(self, **kwargs) -> QImage:
        from_cache = self.get_from_cache(("image", self.name))
        if from_cache is not None:
            return from_cache
        _out = QImage(self.qt_path or str(self.file_path), **kwargs)
        self.store_in_cache(("image", self.name), _out)
        return _out


class MovieRessourceItem(ResourceItem):

    def get_as_movie(self) -> QMovie:
        from_cache = self.get_from_cache(("movie", self.name))
        if from_cache is not None:
            return from_cache
        _out = QMovie(self.qt_path, QByteArray())
        self.store_in_cache(("movie", self.name), _out)
        return _out


_ressource_item_factory_class_table = {ResourceTypus.MISC: MiscResourceItem,
                                       ResourceTypus.PIXMAP: PixmapResourceItem,
                                       ResourceTypus.MOVIE: MovieRessourceItem}


def ressource_item_factory(file_path: str, qt_path: str) -> Union[MiscResourceItem, PixmapResourceItem]:
    file_path = Path(file_path)
    typus = ResourceTypus.from_file_path(file_path)
    return _ressource_item_factory_class_table[typus](file_path=file_path, qt_path=qt_path)


class AllResourceItemsMeta(type):

    def __getattr__(cls, name: str) -> "ResourceItem":
        if any(name.casefold().endswith(f"_{cat}") for cat in cls.categories):
            cat_name = name.split('_')[-1].casefold()
            cls.missing_items[cat_name].add(name)
            return cls.placeholder_image
        raise AttributeError(f"{cls.__name__} has not attribute {name!r}")

# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
Config-Parser that auto-converts values based on an extra configspec file.

currently available:
    - ini, comment-preserving
"""


from .interface import get_config
import sys
from typing import TYPE_CHECKING, TypeVar, TypeGuard, TypeAlias, Final, TypedDict, Generic, Union, Optional, Protocol, runtime_checkable, NoReturn, NewType, Literal, AnyStr, IO, BinaryIO, TextIO, Any
from collections.abc import Hashable, Iterable, Mapping, Sequence, MutableMapping, MutableSet, MutableSequence, Callable, Generator, Collection, Container, Coroutine
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING, Union
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.utility.enums import NamedMetaPath
from gidapptools.utility.kwarg_dict import KwargDict

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.meta_data.meta_info.meta_info_item import MetaInfo
    from gidapptools.meta_data.meta_paths.meta_paths_item import MetaPaths
    meta_items_type = Union[MetaInfo, MetaPaths, object]

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class ConfigKwargs(KwargDict):

    def __init__(self, base_configuration: dict = None, **kwargs) -> None:
        self._path_overwrites: dict[NamedMetaPath, Path] = {}
        self.created_meta_items: dict[str, "meta_items_type"] = {}
        super().__init__(base_defaults=base_configuration, **kwargs)

    def _post_init(self):
        for key, value in self.data.items():
            if NamedMetaPath.is_in_value(key) is True:
                self._path_overwrites[NamedMetaPath(key)] = self.data.pop(key)

    def add_path_overwrite(self, name: Union[NamedMetaPath, str], path: Path) -> None:
        name = name if isinstance(name, NamedMetaPath) else NamedMetaPath(str(name))
        self._path_overwrites[name] = path

    def __setitem__(self, key, item) -> None:
        if key == 'path_overwrites':
            raise KeyError('Not allowed to set path_overwrites this way, use "add_path_overwrite".')
        for meta_item in self.created_meta_items.values():
            if hasattr(meta_item, key):
                raise KeyError('Not allowed to set Meta_item attributes this way, use the meta_item directly.')
        return super().__setitem__(key, item)

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            for meta_item in self.created_meta_items.values():
                if hasattr(meta_item, key):
                    return getattr(meta_item, key)
            if key == 'path_overwrites':
                return self._path_overwrites
            raise

    def __repr__(self) -> str:
        """
        Basic Repr
        !REPLACE!
        """
        return f'{self.__class__.__name__}'
# region [Main_Exec]


if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Any, Union, Callable, Iterable, Optional
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
import PySide6
from PySide6.QtCore import Qt, QAbstractTableModel, QModelIndex, QPersistentModelIndex, QObject

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]

PYSIDE_INDEX_TYPE = Union[QModelIndex, QPersistentModelIndex]


class BasicTableModel(QAbstractTableModel):

    def __init__(self, parent: Optional[QObject] = None) -> None:
        super().__init__(parent=parent)
        self.data_dispatch: dict[Qt.ItemDataRole:Callable[[PYSIDE_INDEX_TYPE], Any]] = {}
        self.header_dispatch: dict[Qt.ItemDataRole:Callable[[int, Qt.Orientation], Any]] = {}
        self.content: Iterable[object] = None

    def data(self, index: PYSIDE_INDEX_TYPE, role: int = ...) -> Any:
        try:
            return self.data_dispatch[role](index)
        except KeyError:
            pass

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = ...) -> Any:
        try:
            return self.header_dispatch[role](section, orientation)
        except KeyError:
            pass

    def rowCount(self, parent: PYSIDE_INDEX_TYPE = None) -> int:
        if self.content is None:
            return 0
        return len(self.content)

# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING, Optional
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
import PySide6
from PySide6.QtWidgets import QStatusBar, QApplication

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import ApplicationNotExistingError

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gidapptools_qt.basics.application import GidQtApplication

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class GidBaseStatusBar(QStatusBar):

    def __init__(self, parent: Optional[PySide6.QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)

    @property
    def app(self) -> "GidQtApplication":
        app = QApplication.instance()
        if app is None:
            raise ApplicationNotExistingError()
        return app
# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import inspect
from typing import Union, Mapping, Callable, Hashable, Iterable
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.enums import MiscEnum

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class dispatch_mark:
    """
    Marks a method as a dispatch function for an instance `BaseDispatchTable`.
    """
    key_attribute_name: str = '_dispatch_key'
    alias_attribute_name: str = '_dispatch_aliases'
    DEFAULT = MiscEnum.DEFAULT

    def __init__(self, dispatch_key: Hashable = None, aliases: Iterable[Hashable] = None):
        self.dispatch_key = dispatch_key
        self.aliases = aliases

    def __call__(self, func: Callable):
        dispatch_key = func.__name__ if self.dispatch_key is None else self.dispatch_key
        aliases = [] if self.aliases is None else self.aliases
        setattr(func, self.key_attribute_name, dispatch_key)
        setattr(func, self.alias_attribute_name, aliases)

        return func


class BaseDispatchTable:
    DEFAULT = MiscEnum.DEFAULT
    mark = dispatch_mark
    extra_dispatch = {}

    def __init__(self,
                 instance: object = None,
                 auto_collect_prefix: str = None,
                 extra_dispatch: Mapping[Hashable, Callable] = None,
                 default_dispatch: Callable = None,
                 key_conversion: Union[Mapping, Callable] = None,
                 aliases: Mapping[Hashable, Hashable] = None) -> None:
        """
        [summary]

        [extended_summary]

        Args:
            instance (object, optional): [description]. Defaults to None.
            auto_collect_prefix (str, optional): [description]. Defaults to None.
            extra_dispatch (Mapping[Hashable, Callable], optional): [description]. Defaults to None.
            default_dispatch (Callable, optional): [description]. Defaults to None.
            key_conversion (Union[Mapping, Callable], optional): [description]. Defaults to None.
            aliases (Mapping[Hashable, Hashable], optional): [description]. Defaults to None.
        """
        self.instance = instance
        self.auto_collect_prefix = auto_collect_prefix
        if extra_dispatch:
            self.extra_dispatch |= extra_dispatch

        self._table: dict[Hashable, Callable] = None
        self.default_dispatch: Callable = default_dispatch
        self._aliases = {} if aliases is None else aliases
        if key_conversion is None:
            self.key_conversion = lambda x: x
        elif isinstance(key_conversion, Mapping):
            self.key_conversion = lambda x: key_conversion.get(x, x)
        else:
            self.key_conversion = key_conversion

    def set_default_dispatch(self, value: Callable) -> None:
        self.default_dispatch = value
        self._table[dispatch_mark.DEFAULT] = value

    def _collect_dispatch_data(self) -> None:
        instance = self if self.instance is None else self.instance
        collected_data = {}
        for meth_name, meth_obj in inspect.getmembers(instance, inspect.ismethod):
            if self.auto_collect_prefix is not None and meth_name.startswith(self.auto_collect_prefix):
                key = self.key_conversion(meth_name.removeprefix(self.auto_collect_prefix))
                collected_data[key] = meth_obj

            key = self.key_conversion(getattr(meth_obj, dispatch_mark.key_attribute_name, None))
            if key in self._aliases:
                raise KeyError(f'An alias cannot be the same as a dispatch key, {key=!r}')
            if key is dispatch_mark.DEFAULT and self.default_dispatch is None:
                self.default_dispatch = meth_obj
            if key is not None:
                collected_data[key] = meth_obj
                aliases = getattr(meth_obj, dispatch_mark.alias_attribute_name)
                for alias in aliases:
                    if alias in self._aliases:
                        raise KeyError(f"Alias {alias!r} already set in aliases.")
                    self._aliases[alias] = key

        self._table = collected_data

    def __getitem__(self, key: Hashable) -> Callable:
        if self._table is None:
            self._collect_dispatch_data()
        combined_table = self._table | self.extra_dispatch
        key = self._aliases.get(key, key)
        return combined_table[key]

    @classmethod
    def add_extra_dispatch(cls, key: Hashable, value: Callable):
        cls.extra_dispatch[key] = value

    def get(self, key: Hashable, default=MiscEnum.NOTHING) -> Callable:
        try:
            return self[key]
        except KeyError:
            return self.default_dispatch if default is MiscEnum.NOTHING else default


# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import re
from typing import Union, ClassVar, Optional
from pathlib import Path
from functools import total_ordering

# * Third Party Imports --------------------------------------------------------------------------------->
import attr

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


__all__ = ["VersionItem"]


def try_convert_int(data: Union[str, int, None]) -> Union[str, int, None]:
    if data is None:
        return None
    if isinstance(data, str) and data in {"", "MISSING"}:
        return None
    try:
        return int(data)
    except ValueError:
        return data


@attr.s(slots=True, auto_attribs=True, auto_detect=True, frozen=True)
@total_ordering
class VersionItem:
    major: int = attr.ib(converter=int)
    minor: int = attr.ib(converter=int)
    patch: int = attr.ib(default=None, converter=try_convert_int)
    extra: Union[str, int] = attr.ib(default=None, converter=try_convert_int)
    version_regex: ClassVar = re.compile(r"(?P<major>\d+)\.(?P<minor>\d+)\.?(?P<patch>\d+|MISSING)?\-?(?P<extra>.*)?")

    def __str__(self) -> str:
        _out = f"{self.major}.{self.minor}"
        if self.patch is not None:
            _out += f".{self.patch}"
        if self.extra is not None:
            _out += f"-{self.extra}"
        return _out

    @classmethod
    def from_string(cls, string: Optional[str]) -> Optional["VersionItem"]:
        if string is None:
            return
        match = cls.version_regex.match(string.strip())
        if match is not None:
            return cls(**match.groupdict())

    def as_tuple(self, include_extra: bool = True) -> tuple[Union[str, int]]:
        if include_extra is False:
            return (self.major, self.minor, self.patch)
        return (self.major, self.minor, self.patch, self.extra)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.as_tuple() == other.as_tuple()

        if isinstance(other, str):
            return False
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if other is None:
            return False
        if isinstance(other, self.__class__):

            if (self.major, self.minor, self.patch) == (other.major, other.minor, other.patch):
                if self.extra is None and other.extra is not None:
                    return True
                return False

            if self.major < other.major:
                return True
            elif self.major == other.major and self.minor < other.minor:
                return True
            elif self.major == other.major and self.minor == other.minor and (self.patch is None and other.patch is not None):
                return True
            elif self.major == other.major and self.minor == other.minor and (all([self.patch is not None, other.patch is not None]) and self.patch < other.patch):
                return True
            elif self.major == other.major and self.minor == other.minor and (all([self.patch is not None, other.patch is not None]) and self.patch == other.patch) and (self.extra is None and other.extra is not None):
                return True
            return False

        if isinstance(other, str):
            return False
        return NotImplemented

# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path
from datetime import datetime, timezone, timedelta
from operator import add
from functools import reduce, cached_property

# * Third Party Imports --------------------------------------------------------------------------------->
import pyparsing as ppa
import pyparsing.common as ppc
from dateutil.tz import gettz

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_parsing.tokens.base_tokens import BaseTokenWithPos

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def microsecond_parse_action(tokens):
    miliseconds_string = tokens[0]
    while len(miliseconds_string) < 6:
        miliseconds_string += "0"
    return int(miliseconds_string)


datetime_format_mapping: dict[str, ppa.ParserElement] = {"%Y": ppa.Regex(r"\d{4}").set_parse_action(ppc.convert_to_integer)("year"),
                                                         "%m": ppa.Regex(r"[01]?\d").set_parse_action(ppc.convert_to_integer)("month"),
                                                         "%d": ppa.Regex(r"[0123]?\d").set_parse_action(ppc.convert_to_integer)("day"),
                                                         "%H": ppa.Regex(r"[012]?\d").set_parse_action(ppc.convert_to_integer)("hour"),
                                                         "%M": ppa.Regex(r"[0-5]?\d").set_parse_action(ppc.convert_to_integer)("minute"),
                                                         "%S": ppa.Regex(r"[0-5]?\d").set_parse_action(ppc.convert_to_integer)("second"),
                                                         "%f": ppa.Regex(r"\d+").set_parse_action(microsecond_parse_action)("microsecond"),
                                                         "%Z": ppa.Combine(ppa.Word(ppa.alphas) + ppa.Optional(ppa.one_of(["+", "-"]) + ppa.Word(ppa.nums) + ppa.Literal(":") + ppa.Word(ppa.nums)))("tzinfo")}


class DateTimeToken(BaseTokenWithPos):

    def __init__(self,
                 start: int,
                 end: int,
                 year: int,
                 month: int,
                 day: int,
                 hour: int,
                 minute: int,
                 second: int,
                 microsecond: int = 0,
                 tzinfo: str = None) -> None:
        super().__init__(start=start, end=end)
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second
        self.microsecond = microsecond
        self.raw_tzinfo = tzinfo

    @classmethod
    def from_parse_action(cls, s, l, t) -> "BaseTokenWithPos":
        data_dict = t[0].as_dict()
        data_dict["start"] = data_dict.pop("locn_start")
        data_dict["end"] = data_dict.pop("locn_end")
        data_dict.pop("value")
        return cls(**data_dict)

    @cached_property
    def tzinfo(self) -> timezone:
        if self.raw_tzinfo is None:
            return None
        if self.raw_tzinfo == "UTC":
            return timezone.utc

        if self.raw_tzinfo.startswith("UTC"):
            offset_part = self.raw_tzinfo.removeprefix("UTC").split(":")[0]
            amount = int(offset_part[1:].strip())
            if offset_part[0] == "-":
                amount = amount * (-1)
            return timezone(timedelta(hours=amount))

        return gettz(self.raw_tzinfo)

    def as_datetime(self) -> datetime:
        return datetime(year=self.year, month=self.month, day=self.day, hour=self.hour, minute=self.minute, second=self.second, microsecond=self.microsecond, tzinfo=self.tzinfo)


def get_grammar_from_dt_format(dt_format: str) -> ppa.ParserElement:
    parts = []
    previous_char = ""
    for char in dt_format:
        if previous_char == "%":
            parts.append(datetime_format_mapping[previous_char + char])
        elif char not in {"%", " "}:
            parts.append(ppa.Literal(char).suppress())
        previous_char = char
    return ppa.locatedExpr(reduce(add, parts)).set_parse_action(DateTimeToken.from_parse_action)


# region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import re
from pathlib import Path

# * Third Party Imports --------------------------------------------------------------------------------->
import pyparsing as pp

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import TrailingCommentError
from gidapptools.gid_config.parser.tokens import Entry, Token, Comment, Section, TokenFactory
from gidapptools.gid_config.parser.grammar import BaseIniGrammar
from gidapptools.general_helper.timing import get_dummy_profile_decorator_in_globals
# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]
get_dummy_profile_decorator_in_globals()
THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class BaseIniParser:
    """
    Parser for text in `ini` format.

    Keeps normal comments, but strips inline comments. Can handle multiline values as long as they are indented. Orphaned Entries (Entries under no section) are not allowed.

    Args:
        grammar_class (BaseIniGrammar, optional): Grammar item to use, must implement the method `get_grammar`. Defaults to `BaseIniGrammar`.
        token_factory (TokenFactory, optional): Factory that transforms pyparsing results into tokens, gets directly passed to the grammar_class. Defaults to `TokenFactory`.
        key_value_separator (str, optional): The string that indicates the separation between a key and its value, gets also used for transforming back into text. Defaults to '='.
        comment_indicator (str, optional): String that indicates the start of a comment, gets also used for transforming back into text. Defaults to '#'.
        remove_all_comments (bool, optional): If the parser should remove all comments and not just inline comments in the preprocessing step. Defaults to False.

    """

    __slots__ = ("comment_indicator", "key_value_separator", "grammar_item", "remove_all_comments", "all_comment_regex", "inline_comment_regex", "grammar")
    all_comment_regex_pattern = r'[ \t]*{comment_indicator}.*'
    inline_comment_regex_pattern = r"(?<=\w)[ \t]*{comment_indicator}[ \t\w]+"

    def __init__(self,
                 grammar_class: BaseIniGrammar = BaseIniGrammar,
                 token_factory: TokenFactory = None,
                 key_value_separator: str = '=',
                 comment_indicator: str = '#',
                 remove_all_comments: bool = False) -> None:

        self.comment_indicator = comment_indicator
        self.key_value_separator = key_value_separator
        self.grammar_item = grammar_class(key_value_separator=self.key_value_separator, comment_indicator=self.comment_indicator, token_factory=token_factory)
        self.remove_all_comments = remove_all_comments
        self.all_comment_regex = re.compile(self.all_comment_regex_pattern.format(comment_indicator=self.comment_indicator))
        self.inline_comment_regex = re.compile(self.inline_comment_regex_pattern.format(comment_indicator=self.comment_indicator))
        self.grammar: pp.ParserElement = None

    def _preprocess_comments(self, text: str) -> str:
        strip_regex = self.all_comment_regex if self.remove_all_comments is True else self.inline_comment_regex
        return strip_regex.sub('', text)

    def _verify(self, text: str) -> str:
        check_text = text.strip()

        if check_text == '':
            return text

        if check_text.splitlines()[-1].strip().startswith(self.comment_indicator):

            raise TrailingCommentError(f'Trailing comments are not allowed with {self.__class__.__name__!r}.')
        return text

    def _pre_process(self, text: str) -> str:
        text = self._preprocess_comments(text)
        text = self._verify(text)
        return text

    def _parse(self, text: str) -> list[Token]:
        temp_comments = []
        last_section = None
        data = []

        def _add_comments_to_token(token) -> None:
            nonlocal temp_comments
            if temp_comments:
                token.comments += temp_comments
                temp_comments.clear()

        def _process_section(token) -> None:
            nonlocal last_section
            _add_comments_to_token(token)
            last_section = token
            data.append(token)

        def _process_entry(token) -> None:
            _add_comments_to_token(token)
            token.key_value_separator = self.key_value_separator
            last_section.add_entry(token)

        def _process_comment(token) -> None:
            nonlocal temp_comments
            token.comment_indicator = self.comment_indicator
            temp_comments.append(token)

        process_table = {Section: _process_section,
                         Entry: _process_entry,
                         Comment: _process_comment}

        for tokens in self.grammar.search_string(text):

            for _token in tokens:

                processor = process_table[type(_token)]
                processor(_token)
        return data

    def parse(self, text: str, **kwargs) -> list[Section]:
        """
        Parses the text into a list of Sections containing the comments and Entries.

        `grammar` is created lazily at this point.

        Args:
            text (str): Text to parse, must be in `ini` format.

        Returns:
            list[Section]: A list of Section, each Section contains the comments for the Section and the Entries.
        """
        if self.grammar is None:
            self.grammar = self.grammar_item.get_grammar(**kwargs)
        text = self._pre_process(text)

        return self._parse(text)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}'
# region [Main_Exec]


if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import sys
import random
from typing import Any, Union, TypeVar, Callable, Hashable, Iterable, Optional, Generator
from logging import Logger
from pathlib import Path
from itertools import tee, filterfalse


# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.enums import MiscEnum

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]

T = TypeVar("T")
_DEFAULT_TYPE = TypeVar("_DEFAULT_TYPE", object, None)


def defaultable_list_pop(in_list: Optional[list[T]], idx: int, default: _DEFAULT_TYPE = None) -> Union[T, _DEFAULT_TYPE]:
    if in_list is None:
        return default
    try:
        return in_list.pop(idx)
    except IndexError:
        return default


T = TypeVar("T")
IN_ITERABLE_TYPE = Iterable[T]
CHECK_FUNC_TYPE = Callable[[object], bool]


class _IterSplitter:
    __slots__ = ("output_type_strategies",)

    def __init__(self) -> None:
        self.output_type_strategies: dict[type, Callable[[IN_ITERABLE_TYPE, CHECK_FUNC_TYPE], tuple[Iterable[T], Iterable[T]]]] = {list: self._output_lists,
                                                                                                                                   set: self._output_sets,
                                                                                                                                   tuple: self._output_tuples,
                                                                                                                                   Generator: self._output_generators}

    def _output_lists(self, in_iterable: Iterable[T], check_func: Callable[[object], bool]) -> tuple[list[T], list[T]]:
        check_true_items = []
        check_false_items = []
        for item, check_result in ((i, check_func(i)) for i in in_iterable):

            if check_result is True:
                check_true_items.append(item)
            elif check_result is False:
                check_false_items.append(item)
            else:
                raise TypeError(f"check_func {check_func!r} returned a non-bool result {check_result!r}")

        return check_true_items, check_false_items

    def _output_sets(self, in_iterable: Iterable[T], check_func: Callable[[object], bool]) -> tuple[set[T], set[T]]:
        check_true_items, check_false_items = self._output_lists(in_iterable=in_iterable, check_func=check_func)
        return set(check_true_items), set(check_false_items)

    def _output_tuples(self, in_iterable: Iterable[T], check_func: Callable[[object], bool]) -> tuple[tuple[T], tuple[T]]:
        check_true_items, check_false_items = self._output_lists(in_iterable=in_iterable, check_func=check_func)
        return tuple(check_true_items), tuple(check_false_items)

    def _output_generators(self, in_iterable: Iterable[T], check_func: Callable[[object], bool]) -> tuple[Generator[T, None, None], Generator[T, None, None]]:
        t1, t2 = tee(in_iterable)
        return (i for i in filter(check_func, t2)), (i for i in filterfalse(check_func, t1))

    def _determine_output_type(self, in_iterable: Iterable[T]) -> Union[type[list], type[tuple], type[set], type[Generator]]:
        if isinstance(in_iterable, set):
            return set

        elif isinstance(in_iterable, tuple):
            return tuple

        elif isinstance(in_iterable, list):
            return list

        elif isinstance(in_iterable, Generator):
            return Generator

    def __call__(self, in_iterable: Iterable[T], check_func: Callable[[object], bool], output_type: Union[type[list], type[tuple], type[set], type[Generator], None] = None):
        if output_type is None:
            output_type = self._determine_output_type(in_iterable=in_iterable)

        strategy = self.output_type_strategies[output_type]
        return strategy(in_iterable=in_iterable, check_func=check_func)


split_iter = _IterSplitter()


def dict_pop_fallback(in_dict: dict, keys: Union[Iterable[Hashable], Hashable], default: Any = None) -> Any:
    for key in keys:
        value = in_dict.pop(key, MiscEnum.NOT_FOUND)
        if value is not MiscEnum.NOT_FOUND:
            return value
    return default


def is_frozen() -> bool:
    return getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS')


def get_all_available_loggers(logger: Logger = None) -> tuple[str]:
    manager = logger.manager if logger is not None else Logger.manager
    names = set(manager.loggerDict)
    return tuple(sorted(names, key=len))


T = TypeVar("T")


def iter_grouped(in_iterable: Iterable[T], group_size: int = 2) -> Generator[tuple[T, ...], None, None]:
    sentinel = object()
    _iterator = iter(in_iterable)
    collected = []
    while True:
        next_item = next(_iterator, sentinel)
        if next_item is sentinel:
            break

        collected.append(next_item)
        if len(collected) == group_size:
            yield tuple(collected)
            collected.clear()

    if len(collected) != 0:
        yield tuple(collected)

# region [Main_Exec]


if __name__ == '__main__':
    # from gidapptools.general_helper.timing import time_execution

    # def checker(in_num: int) -> bool:
    #     return len(in_num) == 2

    # t = [str(random.randint(0, 1_000)) for _ in range(10_000)]

    # with time_execution(also_pretty=True):
    #     a, b = split_iter(t, checker, Generator)
    #     for xx in a:
    #         print(xx)
    for l in get_all_available_loggers():
        print(l)

# endregion [Main_Exec]



"""
[summary]

[extended_summary]
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
from typing import TYPE_CHECKING, Union
from pathlib import Path
from weakref import WeakValueDictionary

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gid_signal.signals.abstract_signal import AbstractSignal

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [AppUserData]


# endregion [AppUserData]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(os.path.dirname(__file__)).absolute()

# endregion [Constants]


class SignalRegistry(WeakValueDictionary):

    def register(self, signal: "AbstractSignal"):
        key = signal.key
        self[key] = signal

    def unregister(self, signal: Union[str, "AbstractSignal"]):
        signal_key = signal.key if isinstance(signal, AbstractSignal) else signal
        self.pop(signal_key)

    def get(self, key: str, default=None) -> "AbstractSignal":
        return super().get(key=key, default=default)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(signals={list(self.data)})"

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(signals={list(self.data)})"

    def clear(self) -> None:
        for key in list(self):
            del self[key]


signal_registry = SignalRegistry()


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import logging.handlers
import re
import logging
from time import perf_counter
from threading import Lock, RLock
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Union, Literal, Callable
from pathlib import Path
from datetime import datetime, timezone
from collections import deque
from logging.handlers import BaseRotatingHandler
from frozendict import frozendict
# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.conversion import human2bytes
from gidapptools.general_helper.regex.datetime_regex import datetime_format_to_regex

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.custom_types import PATH_TYPE
    from gidapptools.gid_logger.records import LOG_RECORD_TYPES

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


DEFAULT_MAX_BYTES = human2bytes("5 mb")


class BaseFileNameTemplate(ABC):

    def __init__(self, base_name: str, suffix: str = ".log") -> None:
        self.base_name = base_name
        self.suffix = suffix

    @abstractmethod
    def format(self) -> str:
        ...

    @abstractmethod
    def make_backup_file_path(self, in_name: str, backup_folder: Path) -> Path:
        ...

    @abstractmethod
    def is_same_kind_log_file(self, other_file: Path) -> bool:
        ...

    @abstractmethod
    def is_backup_log_file(self, other_file: Path) -> bool:
        ...


class TimestampFileNameTemplate(BaseFileNameTemplate):
    format_template: str = "{base_name}_{timestamp}{suffix}"

    def __init__(self, base_name: str, suffix: str = ".log", time_zone: timezone = None, timestamp_format: str = None) -> None:
        super().__init__(base_name=base_name, suffix=suffix)
        self.time_zone = time_zone
        self.timestamp_format = timestamp_format or self._get_default_timestamp_format()
        self.timestamp_regex = datetime_format_to_regex(self.timestamp_format, re.IGNORECASE)

    def _get_default_timestamp_format(self) -> str:
        return "%Y-%m-%d_%H-%M-%S_%Z" if self.time_zone is not None else "%Y-%m-%d_%H-%M-%S"

    def get_timestamp(self) -> str:
        now = datetime.now(tz=self.time_zone)
        return now.strftime(self.timestamp_format)

    def format(self) -> str:
        return self.format_template.format(base_name=self.base_name, timestamp=self.get_timestamp(), suffix=self.suffix)

    def make_backup_file_path(self, in_name: str, backup_folder: Path) -> Path:
        return backup_folder.joinpath(in_name)

    def is_same_kind_log_file(self, other_file: Path) -> bool:
        if other_file.suffix != self.suffix:
            return False

        date_time_match = self.timestamp_regex.search(other_file.stem)
        if not date_time_match:
            return False
        base_name = other_file.stem[:date_time_match.start()].rstrip("_")

        if base_name == self.base_name:
            return True

        return False

    def is_backup_log_file(self, other_file: Path) -> bool:
        return self.is_same_kind_log_file(other_file=other_file)


class GidBaseRotatingFileHandler(BaseRotatingHandler):

    def __init__(self,
                 base_name: str,
                 log_folder: "PATH_TYPE",
                 file_name_template: Union[str, Any] = None,
                 backup_amount_limit: int = 10) -> None:
        self.base_name = base_name
        self.file_name_template = TimestampFileNameTemplate(self.base_name) if file_name_template is None else file_name_template
        self.log_folder = Path(log_folder)
        self.backup_amount_limit = backup_amount_limit
        self.full_file_path: Path = self._construct_full_file_path()
        self.first_record_emited: bool = False
        super().__init__(self.full_file_path, "a", encoding="utf-8", delay=True, errors="ignore")

    def emit(self, record) -> None:
        with self.lock:
            if self.first_record_emited is False:
                self.on_start_rotation()
                self.first_record_emited = True

        return super().emit(record)

    def _construct_full_file_path(self) -> Path:
        name = self.file_name_template.format()
        full_path = self.log_folder.joinpath(name)

        return full_path

    def _get_old_logs(self) -> tuple[Path]:
        def _is_old_log(_in_file: Path) -> bool:
            return _in_file.is_file() and self.file_name_template.is_same_kind_log_file(_in_file)

        _out = tuple(file for file in tuple(self.log_folder.iterdir()) if _is_old_log(file) is True)

        return _out

    def _get_backup_logs(self) -> list[Path]:

        def _is_old_backup(_in_file: Path) -> bool:
            return _in_file.is_file() and self.file_name_template.is_backup_log_file(_in_file)

        _out = sorted((file for file in self.log_folder.iterdir() if _is_old_backup(file)), key=lambda x: x.stat().st_mtime)
        return _out

    def on_start_rotation(self) -> None:
        try:
            self.acquire()
            self.log_folder.mkdir(exist_ok=True, parents=True)
            self.remove_excess_backup_files()

        finally:
            self.release()

    def remove_excess_backup_files(self) -> None:
        if self.backup_amount_limit is None:
            return
        backup_logs = self._get_backup_logs()
        while len(backup_logs) > self.backup_amount_limit:
            to_delete: Path = backup_logs.pop(0)
            to_delete.unlink(missing_ok=True)

    def shouldRollover(self, record: logging.LogRecord) -> bool:
        return False


class GidBaseStreamHandler(logging.StreamHandler):

    def __init__(self, stream=None):
        super().__init__(stream=stream)


LOG_DEQUE_TYPE = deque["LOG_RECORD_TYPES"]


class GidStoringHandler(logging.Handler):

    def __init__(self, max_storage_size: int = None) -> None:
        super().__init__()
        self.emit_lock = RLock()
        self._callbacks: dict[str, set[Callable[["LOG_RECORD_TYPES" | None], None]]] = {"ALL": set(),
                                                                                        "DEBUG": set(),
                                                                                        "INFO": set(),
                                                                                        "WARNING": set(),
                                                                                        "CRITICAL": set(),
                                                                                        "ERROR": set()}
        self.debug_messages: "LOG_DEQUE_TYPE" = deque(maxlen=max_storage_size)
        self.info_messages: "LOG_DEQUE_TYPE" = deque(maxlen=max_storage_size)
        self.warning_messages: "LOG_DEQUE_TYPE" = deque(maxlen=max_storage_size)
        self.critical_messages: "LOG_DEQUE_TYPE" = deque(maxlen=max_storage_size)
        self.error_messages: "LOG_DEQUE_TYPE" = deque(maxlen=max_storage_size)
        self.other_messages: "LOG_DEQUE_TYPE" = deque(maxlen=max_storage_size)

        self._all_messages: "LOG_DEQUE_TYPE" = deque(maxlen=max_storage_size)

        self.table = frozendict({'CRITICAL': self.critical_messages,
                                 'FATAL': self.critical_messages,
                                 'ERROR': self.error_messages,
                                 'WARN': self.warning_messages,
                                 'WARNING': self.warning_messages,
                                 'INFO': self.info_messages,
                                 'DEBUG': self.debug_messages,
                                 "OTHER": self.other_messages,
                                 "ALL": self._all_messages})

    @property
    def all_deques(self) -> tuple["LOG_DEQUE_TYPE"]:
        # dupicates = {"FATAL", "WARN"}
        return tuple({k: v for k, v in self.table.items() if k not in {"FATAL", "WARN"}}.values())
        # _out = []
        # for table in self.table.values():
        #     if table not in _out:
        #         _out.append(table)

        # return tuple(_out)

    def add_callback(self, typus: Literal["ALL", "DEBUG", "INFO", "WARNING", "CRITICAL", "ERROR"], callback: Callable[["LOG_RECORD_TYPES" | None], None]):
        callback_list = self._callbacks[typus]

        callback_list.add(callback)

    def remove_callback(self, typus: Literal["ALL", "DEBUG", "INFO", "WARNING", "CRITICAL", "ERROR"], callback: Callable[["LOG_RECORD_TYPES" | None], None]):
        try:
            self._callbacks[typus].remove(callback)
        except KeyError:
            pass

    def send_to_callbacks(self, typus: Literal["ALL", "DEBUG", "INFO", "WARNING", "CRITICAL", "ERROR"], record: "LOG_RECORD_TYPES" | None):
        for callback in self._callbacks.get(typus, []):
            callback(record)

    def set_max_storage_size(self, max_storage_size: int = None):
        for store in self.table.values():
            store.maxlen = max_storage_size

    def emit(self, record: "LOG_RECORD_TYPES") -> None:
        with self.emit_lock:
            target = self.table.get(record.levelname, self.other_messages)

            target.append(record)

            self._all_messages.append(record)

            self.send_to_callbacks(typus="ALL", record=record)

            typus = record.levelname.upper()
            if typus == "WARN":
                typus = "WARNING"

            if typus == "FATAL":
                typus = "CRITICAL"

            self.send_to_callbacks(typus=typus, record=record)

    def get_stored_messages(self) -> dict[str, tuple["LOG_RECORD_TYPES"]]:

        return {k: tuple(v) for k, v in self.table.items() if k not in {"FATAL", "WARN"}}
        # _out = {}
        # for level, store in self.table.items():
        #     if level == "FATAL":
        #         level = "CRITICAL"
        #     elif level == "WARN":
        #         level = "WARNING"
        #     _out[level] = tuple(store)

        # return _out

    def get_all_messages(self, formatted: bool = False) -> tuple["LOG_RECORD_TYPES"]:
        with self.emit_lock:
            all_messages = tuple(self._all_messages)
            if formatted is True:
                list(self.format(r) for r in all_messages)

            return all_messages

    def get_formated_messages(self) -> dict[str, tuple[str]]:
        _out = {}
        for level, store in self.table.items():
            if level == "FATAL":
                level = "CRITICAL"
            elif level == "WARN":
                level = "WARNING"
            _out[level] = tuple(self.format(r) for r in store)
        return _out

    def __len__(self) -> int:
        with self.emit_lock:
            return len(self._all_messages)

    def clear(self, typus: None | Literal["debug", "info", "warning", "critical", "error", "other"] = None):
        if typus is not None:
            _deque: "LOG_DEQUE_TYPE" = getattr(self, f"{typus.casefold()}_messages")
            records = tuple(_deque)
            _deque.clear()
            for record in records:
                self._all_messages.remove(record)

            self.send_to_callbacks(typus=typus.upper(), record=None)

        else:
            for _deque in (self.debug_messages, self.info_messages, self.warning_messages, self.critical_messages, self.error_messages, self.other_messages, self._all_messages):
                _deque.clear()

            self.send_to_callbacks(typus="ALL", record=None)

# region [Main_Exec]


if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path

# * Third Party Imports --------------------------------------------------------------------------------->
from frozendict import frozendict

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]

STD_DEFAULT_PRAGMAS = frozendict({
    "wal_autocheckpoint": 1_000,
    "auto_vacuum": 2,
    "cache_size": -1 * 256_000,  # 128mb
    "journal_mode": 'WAL',
    "synchronous": 0,
    "ignore_check_constraints": 0,
    "foreign_keys": 1,
    "journal_size_limit": 786432000,  # 750mb
    "page_size": 8192,
    "analysis_limit": 100_000,
    "case_sensitive_like": False,
    "threads": 8,
    # "temp_store": "MEMORY",
    # "mmap_size": 30_000_000_000,
})

STD_DEFAULT_EXTENSIONS = frozendict({"c_extensions": True,
                                     "rank_functions": False,
                                     "hash_functions": True,
                                     "json_contains": True,
                                     "bloomfilter": False,
                                     "regexp_function": True})


MEMORY_DB_PATH = ":memory:"

# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import logging.handlers
from typing import Optional, NamedTuple, Literal, Iterable
from pathlib import Path
import logging
import math
from functools import partial
import re
# * Third Party Imports --------------------------------------------------------------------------------->
from pyparsing.exceptions import ParseBaseException

# * Qt Imports --------------------------------------------------------------------------------------->
import PySide6
from PySide6 import QtGui, QtWidgets, QtCore
from PySide6.QtGui import QFont, QColor, QPalette, QBrush, QMouseEvent, QTextCharFormat, QSyntaxHighlighter, QPainter
from PySide6.QtCore import Qt, Slot, Signal, QSize
from PySide6.QtWidgets import (QLabel, QWidget, QStyle, QApplication, QFrame, QVBoxLayout, QStyleOption, QStyleOptionViewItem, QComboBox, QSizePolicy, QGroupBox,
                               QTextEdit, QFormLayout, QStyledItemDelegate, QGridLayout, QRadioButton, QCheckBox, QHBoxLayout, QPushButton, QSpacerItem, QTableView,
                               QTableWidget, QTableWidgetItem, QAbstractItemView)

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_logger.logger import get_main_logger
from gidapptools.gid_logger.handler import GidStoringHandler
from gidapptools.general_helper.conversion import bytes2human
from gidapptools.general_helper.string_helper import StringCaseConverter, StringCase
from gidapptools.gid_parsing.py_log_parsing import GeneralGrammar

import sys
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]


THIS_FILE_DIR = Path(__file__).parent.absolute()


# endregion [Constants]


class AppLogHighlighter(QSyntaxHighlighter):
    grammar = GeneralGrammar()
    level_regex = re.compile(r"(?P<level>(DEBUG)|(INFO)|(WARN(ING)?)|(CRITICAL)|(ERROR))")
    line_number_regex = re.compile(r"\| *(?P<line_number>\d+) *\|")

    def __init__(self, parent=None):
        super().__init__(parent)
        self.formats: dict[str, QTextCharFormat] = {}
        self.setup_formats()

    def setup_formats(self):
        self.base_format = QTextCharFormat()

        date_format = QTextCharFormat()
        date_format.setForeground(QColor(0, 0, 0, 255))
        date_format.setTextOutline(QColor(255, 255, 255, 50))
        date_format.setFontWeight(1000)
        self.formats["time_stamp"] = date_format

        line_number_format = QTextCharFormat()
        line_number_format.setForeground(QColor(0, 0, 0, 255))
        line_number_format.setTextOutline(QColor(255, 255, 255, 50))
        line_number_format.setFontWeight(1000)
        self.formats["line_number"] = line_number_format

        level_format = QTextCharFormat()
        level_format.setFontWeight(1000)
        level_format.setFontUnderline(True)
        self.formats["level"] = level_format

        debug_format = QTextCharFormat()
        debug_format.setBackground(QColor(144, 238, 144, 75))
        self.formats["debug"] = debug_format

        info_format = QTextCharFormat()
        info_format.setBackground(QColor(173, 216, 230, 75))
        self.formats["info"] = info_format

        critical_format = QTextCharFormat()
        critical_format.setBackground(QColor(255, 165, 0, 75))
        self.formats["critical"] = critical_format

        error_format = QTextCharFormat()
        error_format.setBackground(QColor(220, 20, 60, 100))
        self.formats["error"] = error_format

        message_format = QTextCharFormat()
        message_format.setFontWeight(1000)
        self.formats["message"] = message_format

        thread_format = QTextCharFormat()
        self.formats["thread"] = thread_format

        module_format = QTextCharFormat()
        self.formats["module"] = module_format

        function_format = QTextCharFormat()
        self.formats["function"] = function_format

    def highlightBlock(self, text: str) -> None:

        # try:
        #     tokens = self.grammar(text)
        #     background_fmt = self.formats[tokens["level"].log_level.casefold()]
        #     self.setFormat(0, len(text), background_fmt)
        #     for name, token in tokens.items():
        #         fmt = self.formats.get(name, self.base_format)
        #         fmt.setBackground(background_fmt.background())
        #         self.setFormat(token.start, token.span, fmt)

        # except ParseBaseException as e:
        # print(f"{e=} | {e.args=}", flush=True)
        if match := self.line_number_regex.search(text):
            self.setFormat(match.start("line_number"), match.end("line_number") - match.start("line_number"), self.formats["line_number"])

        if match := self.level_regex.search(text):
            self.setFormat(0, len(text), self.formats.get(match.group("level").casefold(), self.base_format))

    # def highlightBlock(self, text: str) -> None:
    #     try:
    #         backgrounds = {"debug": QColor(0, 200, 0, 50),
    #                        "info": QColor(0, 0, 255, 50),
    #                        "critical": QColor(255, 200, 0, 50),
    #                        "error": QColor(255, 0, 0, 100)}
    #         if not text.strip():
    #             return
    #         parts = text.split("|")

    #         level_part = parts[2]

    #         start = text.find(level_part)
    #         background = backgrounds.get(level_part.strip().casefold(), QColor(0, 0, 0, 0))
    #         format_item = QTextCharFormat()
    #         format_item.setBackground(background)
    #         self.setFormat(0, text.find("||-->"), format_item)

    #         date_part = parts[0].strip()
    #         start = text.find(date_part)
    #         fmt = self.formats["date"]
    #         fmt.setBackground(background)
    #         self.setFormat(start, len(date_part), fmt)

    #         line_number_part = parts[1].strip()
    #         start = text.find(line_number_part)
    #         fmt = self.formats["line_number"]
    #         fmt.setBackground(background)
    #         self.setFormat(start, len(line_number_part), fmt)

    #         message_start = text.find("||--> ") + 6
    #         self.setFormat(message_start, len(text) - message_start, self.formats["message"])
    #     except IndexError:
    #         pass


class MetaBox(QGroupBox):

    def __init__(self, log_file: Path, parent=None):
        super().__init__(parent=parent)
        self.log_file = log_file
        self.setLayout(QFormLayout())

        self.setTitle("Meta Data")
        self.widgets = {}

        self.setup()

    def _gather_meta_data(self) -> dict[str, str]:
        data = {}
        data["Name"] = self.log_file.name
        data["Path"] = self.log_file.as_posix()
        data["Size"] = bytes2human(self.log_file.stat().st_size)
        data["Lines"] = len(self.log_file.read_text(encoding='utf-8', errors='ignore').splitlines())

        return data

    def setup(self):
        for k, v in self._gather_meta_data().items():
            value_widget = QLabel()
            value_widget.setText(str(v))
            self.layout.addRow(k, value_widget)
            self.widgets[k] = value_widget

    @property
    def layout(self) -> QFormLayout:
        return super().layout()

    def update_size(self):
        size_widget = self.widgets["Size"]
        size_widget.setText(bytes2human(self.log_file.stat().st_size))
        size_widget.repaint()

        line_amount_widget = self.widgets["Lines"]
        line_amount_widget.setText(str(len(self.log_file.read_text(encoding='utf-8', errors='ignore').splitlines())))
        line_amount_widget.repaint()


class FileAppLogViewer(QWidget):

    def __init__(self, log_file: Path, parent=None) -> None:
        super().__init__(parent=parent)
        self.log_file = Path(log_file).resolve()
        self.file_size = None
        self.timer_id = None

    def setup(self) -> "FileAppLogViewer":
        self.setLayout(QGridLayout())
        self.setWindowTitle("Application Log")

        self.setup_widgets()
        self.set_content()
        # self.file_watcher = QFileSystemWatcher(self)
        # self.file_watcher.addPath(str(self.log_file))
        # self.file_watcher.fileChanged.connect(self.set_content)
        # self.file_watcher.fileChanged.connect(self.set_content)
        self.timer_id = self.startTimer(500, Qt.CoarseTimer)
        return self

    def setup_widgets(self):
        self.meta_box = MetaBox(self.log_file)
        self.layout.addWidget(self.meta_box)

        self.text_widget = QTextEdit(self)
        self.text_widget.setReadOnly(True)
        self.text_widget.setLineWrapMode(QTextEdit.NoWrap)
        font: QFont = self.text_widget.font()
        font.setStyleHint(QFont.Monospace)
        font.setFamily("Consolas")
        self.text_widget.setFont(font)

        self.layout.addWidget(self.text_widget)
        self.highlighter = AppLogHighlighter()
        self.highlighter.setDocument(self.text_widget.document())

    def set_content(self):
        self.file_size = self.log_file.stat().st_size
        content = self.log_file.read_text(encoding='utf-8', errors='ignore')
        self.text_widget.setPlainText(content)
        self.resize_to_content()
        self.text_widget.verticalScrollBar().setValue(self.text_widget.verticalScrollBar().maximum())
        self.text_widget.update()
        self.text_widget.repaint()

    def resize_to_content(self):
        height = self.size().height()
        width = min(2000, self.text_widget.document().size().toSize().width())
        self.resize(width, height)

    def check_file(self):
        if self.log_file.stat().st_size != self.file_size:
            self.set_content()
            self.meta_box.update_size()

    def timerEvent(self, event: PySide6.QtCore.QTimerEvent) -> None:
        self.check_file()
        return super().timerEvent(event)

    @property
    def layout(self) -> QGridLayout:
        return super().layout()

    def closeEvent(self, event: PySide6.QtGui.QCloseEvent) -> None:
        if self.timer_id is not None:
            self.killTimer(self.timer_id)
        event.accept()


class LogLevelSelector(QGroupBox):
    level_selection_changed = Signal(tuple)

    def __init__(self, parent: QWidget = None) -> None:
        super().__init__(parent)
        self.setLayout(QVBoxLayout())
        self._level_widgets: dict[str, QCheckBox] = {}

    @property
    def layout(self) -> QVBoxLayout:
        return super().layout()

    @property
    def all_levels_checked(self) -> bool:
        return all(_level_widget.isChecked() for _level_widget in self._level_widgets.values())

    def get_activated_level_names(self) -> tuple[Literal["DEBUG", "INFO", "WARNING", "CRITICAL", "ERROR"]]:
        return tuple(name for name, checkbox in self._level_widgets.items() if checkbox.isChecked())

    def _setup_level_widgets(self):
        for level_name in ("CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"):
            _level_check_box = QCheckBox(level_name, self)
            _level_check_box.checkStateChanged.connect(partial(self.on_level_checkStateChanged, level_name=level_name))
            self._level_widgets[level_name] = _level_check_box
            self.layout.addWidget(_level_check_box)

    def on_level_checkStateChanged(self, state: Qt.CheckState, level_name: str):
        self.level_selection_changed.emit(self.get_activated_level_names())

    def set_all_checked_states(self, checked: bool):
        for _level_check_box in self._level_widgets.values():
            _level_check_box.setChecked(checked)

    def invert_all_checked_states(self):
        for _level_check_box in self._level_widgets.values():
            _level_check_box.setChecked(not _level_check_box.isChecked())

    def setup(self) -> Self:
        self.setTitle("Log Levels")
        self._setup_level_widgets()
        self.set_all_checked_states(True)

        return self

    def mousePressEvent(self, event: QMouseEvent) -> None:
        if event.button() == Qt.MouseButton.RightButton and any(_level_widget.geometry().contains(event.position().toPoint()) for _level_widget in self._level_widgets.values()):

            if self.all_levels_checked is True:
                self.set_all_checked_states(False)
            else:
                self.set_all_checked_states(True)
            event.accept()

        super().mousePressEvent(event)


class StoredAppLogViewer(QWidget):
    closed_signal = Signal()

    def __init__(self, logger: logging.Logger, parent: Optional[PySide6.QtWidgets.QWidget] = None, storage_handler: GidStoringHandler = None) -> None:
        super().__init__(parent)
        self.logger = logger
        if storage_handler is None:
            self.storage_handler: GidStoringHandler = self._try_find_storage_handler()
        else:
            self.storage_handler: GidStoringHandler = storage_handler
        self.last_len = 0
        self.last_active_level_names: frozenset[Literal["DEBUG", "INFO", "WARNING", "CRITICAL", "ERROR"]] = frozenset(["DEBUG", "INFO", "WARNING", "CRITICAL", "ERROR"])
        self.timer_id = None

    def _try_find_storage_handler(self) -> logging.Handler:
        logger = self.logger

        while len(logger.handlers) <= 0:
            logger = logging.getLogger(logger.name.rsplit(".", 1)[0])

        all_handlers = list(logger.handlers)

        for _handler in tuple(all_handlers):
            if isinstance(_handler, logging.handlers.QueueHandler):
                for qued_listener in _handler.listener:
                    for qued_handler in qued_listener.handlers:
                        all_handlers.append(qued_handler)

        for handler in tuple(all_handlers):
            if isinstance(handler, GidStoringHandler):
                return handler

        raise ValueError("no storing handler found.")

    def setup(self) -> Self:
        self.setLayout(QGridLayout())
        self.setWindowTitle("Application Log")

        self.setup_widgets()
        self.gather_content()

        self.timer_id = self.startTimer(int(0.5 * 1000), Qt.CoarseTimer)

        return self

    def setup_widgets(self):
        self.text_widget = QTextEdit(self)
        self.text_widget.setReadOnly(True)
        self.text_widget.setLineWrapMode(QTextEdit.NoWrap)
        font: QFont = self.text_widget.font()
        font.setStyleHint(QFont.Monospace)
        font.setFamily("Consolas")
        font.setPointSizeF(font.pointSizeF() * 1.25)
        self.text_widget.setFont(font)
        self.highlighter = AppLogHighlighter()
        self.highlighter.setDocument(self.text_widget.document())
        self.layout.addWidget(self.text_widget)

        self.clear_button = QPushButton("Clear")
        self.clear_button.pressed.connect(self.on_clear_pressed)
        self.layout.addWidget(self.clear_button)

    @ Slot()
    def on_level_selection_changed(self, level_text: str):
        ...

    @ Slot()
    def on_clear_pressed(self, checked: bool = False):
        self.storage_handler.clear()
        self.gather_content()

    @ property
    def layout(self) -> QGridLayout:
        return super().layout()

    def gather_content(self):
        if len(self.storage_handler) != self.last_len:
            all_messages = []
            for message_tuple in self.storage_handler.get_stored_messages().values():
                for raw_message in message_tuple:
                    all_messages.append(raw_message)
            all_messages = sorted(all_messages, key=lambda x: (x.created, x.msecs))
            text = ""
            for msg in all_messages:
                text += self.storage_handler.format(msg) + '\n'
            h_scroll_value = self.text_widget.horizontalScrollBar().value()
            self.text_widget.setPlainText(text)
            self.text_widget.verticalScrollBar().setValue(self.text_widget.verticalScrollBar().maximum())
            self.text_widget.horizontalScrollBar().setValue(h_scroll_value)

            self.last_len = len(self.storage_handler)

    def timerEvent(self, event: PySide6.QtCore.QTimerEvent) -> None:
        self.gather_content()
        return super().timerEvent(event)

    def closeEvent(self, event: PySide6.QtGui.QCloseEvent) -> None:
        if self.timer_id is not None:
            self.killTimer(self.timer_id)

        self.closed_signal.emit()
        event.accept()


class HighlightDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.doc = QtGui.QTextDocument(self.parent())
        self._highlight_data: list[tuple[QtCore.QRegularExpression, QtGui.QTextCharFormat]] = []

    def paint(self, painter, option, index):
        painter.save()
        options = QtWidgets.QStyleOptionViewItem(option)
        self.initStyleOption(options, index)
        self.doc.setDefaultFont(options.font)
        self.doc.setPlainText(options.text)

        palette: QPalette = options.palette

        bkgrnd_color = index.data(Qt.ItemDataRole.BackgroundRole)
        if bkgrnd_color:

            palette.setColor(QPalette.ColorRole.Base, bkgrnd_color.color())

        foreground_color = index.data(Qt.ItemDataRole.ForegroundRole)
        if foreground_color:
            palette.setColor(QPalette.ColorRole.Text, foreground_color.color())

        self.apply_highlight()
        options.text = ""
        style = QtWidgets.QApplication.style() if options.widget is None else options.widget.style()
        style.drawControl(QtWidgets.QStyle.CE_ItemViewItem, options, painter)

        ctx = QtGui.QAbstractTextDocumentLayout.PaintContext()
        if option.state & QtWidgets.QStyle.State_Selected:
            ctx.palette.setColor(QtGui.QPalette.Text, option.palette.color(
                QtGui.QPalette.Active, QtGui.QPalette.HighlightedText))
        else:
            ctx.palette.setColor(QtGui.QPalette.Text, option.palette.color(
                QtGui.QPalette.Active, QtGui.QPalette.Text))

        textRect = style.subElementRect(
            QtWidgets.QStyle.SE_ItemViewItemText, options)

        if index.column() != 0:
            textRect.adjust(5, 0, 0, 0)

        the_constant = 4
        margin = (option.rect.height() - options.fontMetrics.height()) // 2
        margin = margin - the_constant
        textRect.setTop(textRect.top() + margin)

        painter.translate(textRect.topLeft())
        painter.setClipRect(textRect.translated(-textRect.topLeft()))
        self.doc.documentLayout().draw(painter, ctx)

        painter.restore()

    def apply_highlight(self):
        for regex, fmt in self._highlight_data:
            cursor = QtGui.QTextCursor(self.doc)
            cursor.beginEditBlock()
            highlightCursor = QtGui.QTextCursor(self.doc)
            while not highlightCursor.isNull() and not highlightCursor.atEnd():
                highlightCursor = self.doc.find(regex, highlightCursor)
                if not highlightCursor.isNull():
                    highlightCursor.mergeCharFormat(fmt)
            cursor.endEditBlock()

    def add_format_item(self, regex: QtCore.QRegularExpression, fmt: QtGui.QTextFormat) -> None:
        self._highlight_data.append((regex, fmt))

    def copy(self, parent=None) -> "HighlightDelegate":
        new_highlight_delegate = self.__class__(parent)

        new_highlight_delegate._highlight_data = list(self._highlight_data)

        return new_highlight_delegate


class ColumnDataItem:

    __slots__ = ("_attr_name",
                 "_display_name",
                 "_alignment",
                 "_delegate")

    def __init__(self,
                 attr_name: str,
                 display_name: str | None = None,
                 alignment: Qt.AlignmentFlag = None,
                 delegate: QStyledItemDelegate | None = None) -> None:
        self._attr_name = attr_name
        self._display_name = display_name if display_name is not None else self._get_auto_display_name()
        self._alignment = alignment
        self._delegate = delegate

    @ property
    def attr_name(self) -> str:
        return self._attr_name

    @ property
    def display_name(self) -> str:
        return self._display_name

    @ property
    def alignment(self) -> Qt.AlignmentFlag | None:
        return self._alignment

    @ property
    def delegate(self) -> QStyledItemDelegate | None:
        return self._delegate

    def _get_auto_display_name(self) -> str:
        return StringCaseConverter.convert_to(self.attr_name, StringCase.TITLE)


def get_date_style_delegate(parent=None):
    highlight_delegate = HighlightDelegate(parent)

    date_regex = QtCore.QRegularExpression(r"\d+\-\d+\-\d+")

    date_fmt = QtGui.QTextCharFormat()

    date_fmt.setForeground(QApplication.palette().dark())
    date_fmt.setFontWeight(500)

    highlight_delegate.add_format_item(date_regex, date_fmt)

    time_regex = QtCore.QRegularExpression(r"\d+\:\d+\:\d+\.\d*")

    time_fmt = QtGui.QTextCharFormat()
    QApplication.style
    time_fmt.setForeground(QApplication.palette().placeholderText())

    time_fmt.setFontUnderline(True)

    highlight_delegate.add_format_item(time_regex, time_fmt)

    return highlight_delegate


# _DATE_STYLE_DELEGATE = get_date_style_delegate()


def get_message_style_delegate(parent=None):
    highlight_delegate = HighlightDelegate(parent)

    string_regex = QtCore.QRegularExpression(r"(?P<quotes>\'|\").*?(?P=quotes)")

    string_fmt = QtGui.QTextCharFormat()

    string_fmt.setForeground(QApplication.palette().link())

    string_fmt.setFontItalic(True)

    highlight_delegate.add_format_item(string_regex, string_fmt)

    return highlight_delegate


class StoredAppLogTableViewer(StoredAppLogViewer):

    def __init__(self,
                 logger: logging.Logger,
                 parent=None,
                 storage_handler: GidStoringHandler = None) -> None:
        super().__init__(parent=parent, logger=logger, storage_handler=storage_handler)
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.MinimumExpanding)

        self.error_background_color = QColor(Qt.GlobalColor.red)
        self.error_background_color.setAlpha(50)

        self.critical_background_color = QColor.fromRgb(255, 165, 0)
        self.critical_background_color.setAlpha(50)

        self.warning_background_color = QColor(Qt.GlobalColor.blue)
        self.warning_background_color.setAlpha(50)

        self.debug_background_color = QColor(Qt.GlobalColor.gray)
        self.debug_background_color.setAlpha(50)

        self.debug_foreground_color = QColor(Qt.GlobalColor.darkGray)

    def _get_color_map(self):
        return {"background": {"error": ...,
                               "critical": ...,
                               "warning": ...,
                               "info": ...,
                               "debug": ...},
                "foregorund": {"error": ...}}

    def _setup_level_select_widget(self) -> LogLevelSelector:
        level_select_widget = LogLevelSelector(self).setup()

        level_select_widget.level_selection_changed.connect(self.on_level_selection_changed)
        return level_select_widget

    def setup_widgets(self):
        self.level_select_widget = self._setup_level_select_widget()
        self.layout.addWidget(self.level_select_widget, 0, 0, 2, 1)
        self.table_widget = QTableWidget(self)

        self.column_data = (ColumnDataItem(attr_name="asctime"),
                            ColumnDataItem(attr_name="levelname", display_name="Level Name", alignment=Qt.AlignmentFlag.AlignCenter),
                            ColumnDataItem(attr_name="name", alignment=Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter),
                            ColumnDataItem(attr_name="lineno", display_name="Line Number", alignment=Qt.AlignmentFlag.AlignCenter),
                            ColumnDataItem(attr_name="funcName", alignment=Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter),
                            ColumnDataItem(attr_name="message", alignment=Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter))

        self.table_widget.horizontalHeader().setStretchLastSection(True)
        self.table_widget.setEditTriggers(QTableWidget.EditTrigger.NoEditTriggers)
        self.table_widget.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.table_widget.setColumnCount(len(self.column_data))
        self.table_widget.setHorizontalHeaderLabels([item.display_name for item in self.column_data])
        self.table_widget.horizontalHeader().setMinimumSectionSize(100)
        self.table_widget.verticalHeader().setMinimumSectionSize(25)
        self.table_widget.setWordWrap(False)
        self.table_widget.setTextElideMode(Qt.TextElideMode.ElideNone)
        for column, column_data in enumerate(self.column_data):
            if column_data.delegate is not None:
                column_data.delegate.setParent(self.table_widget)
                self.table_widget.setItemDelegateForColumn(column, column_data.delegate)

        font: QFont = self.table_widget.font()
        font.setStyleHint(QFont.Monospace)
        font.setFamily("Consolas")
        font.setPointSizeF(font.pointSizeF() * 1.25)
        self.table_widget.setFont(font)

        self.layout.addWidget(self.table_widget, 0, 1, 4, 9)

        self.clear_button = QPushButton("Clear Current Logs")
        self.clear_button.pressed.connect(self.on_clear_pressed)
        self.layout.addWidget(self.clear_button, 3, 0, 1, 1)

    @ Slot()
    def on_clear_pressed(self, checked: bool = False):
        current_selecte_level_names = self.level_select_widget.get_activated_level_names()
        for _level_name in current_selecte_level_names:
            self.storage_handler.clear(_level_name.casefold())
        self.gather_content()

    @ Slot()
    def on_level_selection_changed(self, active_level_names: Iterable[Literal["DEBUG", "INFO", "WARNING", "CRITICAL", "ERROR"]]):
        self.gather_content(active_level_names)

    def gather_content(self, active_level_names: Iterable[Literal["DEBUG", "INFO", "WARNING", "CRITICAL", "ERROR"]] | None = None):
        active_level_names = frozenset(active_level_names if active_level_names is not None else self.last_active_level_names)
        all_relevant_messages = [r for r in self.storage_handler.get_all_messages(True) if r.levelname.upper() in active_level_names]

        if self.last_active_level_names == active_level_names and len(all_relevant_messages) == self.table_widget.rowCount():
            return

        self.table_widget.clearContents()

        h_scroll_value = self.table_widget.horizontalScrollBar().value()

        self.table_widget.setRowCount(len(all_relevant_messages))

        for row, msg in enumerate(all_relevant_messages):

            for column, column_data in enumerate(self.column_data):
                if column_data.attr_name == "asctime":
                    item = QTableWidgetItem(self.storage_handler.formatter.formatTime(msg))

                elif column_data.attr_name == "message":
                    item = QTableWidgetItem(msg.message + (f"\n{msg.exc_text}" if msg.exc_text else ""))
                else:
                    value = getattr(msg, column_data.attr_name)

                    text = "" if value is None else str(value)

                    item = QTableWidgetItem(text)

                if column_data.alignment is not None:
                    item.setTextAlignment(column_data.alignment)

                _pathname = Path(msg.pathname).resolve()
                if msg.exc_text:
                    item.setToolTip(f"{_pathname.as_posix()!r}\n\n{msg.exc_text}")

                else:
                    item.setToolTip(f"{_pathname.as_posix()!r}")

                match msg.levelname.casefold():
                    case "error":

                        item.setBackground(self.error_background_color)

                    case "critical":

                        item.setBackground(self.critical_background_color)

                    case "warning" | "warn":

                        item.setBackground(self.warning_background_color)

                    case "debug":

                        item.setBackground(self.debug_background_color)

                        item.setForeground(self.debug_foreground_color)

                    case "info":
                        pass

                self.table_widget.setItem(row, column, item)
            if msg.exc_text:
                self.table_widget.resizeRowToContents(row)
                # self.table_widget.setRowHeight(row, self.table_widget.fontMetrics().h)

        self.table_widget.verticalScrollBar().setValue(self.table_widget.verticalScrollBar().maximum())
        self.table_widget.horizontalScrollBar().setValue(h_scroll_value)

        self.last_len = self.table_widget.rowCount()
        self.last_active_level_names = frozenset(active_level_names)

        if self.table_widget.rowCount() > 0:
            self.table_widget.resizeColumnsToContents()
            self.table_widget.resizeRowsToContents()


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import inspect
from abc import ABCMeta
from typing import Any, Union, Iterable, Optional
from pathlib import Path
from collections import UserDict

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.enums import MiscEnum

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class PostInitMeta(ABCMeta):
    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        instance = super(PostInitMeta, cls).__call__(*args, **kwargs)
        if hasattr(instance, "_post_init"):
            instance._post_init()
        return instance


class KwargDict(UserDict, metaclass=PostInitMeta):
    # pylint: disable=useless-super-delegation
    def __init__(self, base_defaults: dict = None, **kwargs) -> None:
        super().__init__(base_defaults, **kwargs)

    def get_kwargs_for(self, target_class: object, defaults: dict[str, Any] = None, exclude: Iterable[str] = None, overwrites: dict[str, Any] = None) -> Optional[dict[str, Any]]:
        defaults = {} if defaults is None else defaults
        overwrites = {} if overwrites is None else overwrites
        exclude = set() if exclude is None else set(exclude)
        kwarg_names = [name for name, obj in inspect.signature(target_class).parameters.items() if obj.kind is not obj.VAR_KEYWORD]

        _kwargs = {}
        for name in kwarg_names:
            if name in exclude:
                continue
            if name in overwrites:
                value = overwrites.get(name)
            else:
                value = self.get(name, defaults.get(name, MiscEnum.NOTHING))

            if value is MiscEnum.NOTHING or value is MiscEnum.OPTIONAL:
                continue
            _kwargs[name] = value
        return _kwargs

    def get_many(self, keys: Union[list[str], dict[str, Any]]) -> dict[str, Any]:
        keys = keys if isinstance(keys, dict) else {key: None for key in keys}
        result = {}
        for key in keys:
            result[key] = self.get(key, keys.get(key, None))
        return result


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import re
from pathlib import Path

# * Third Party Imports --------------------------------------------------------------------------------->
from frozendict import frozendict

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]

DATETIME_FORMAT_REGEX_MAPPING: frozendict[str, str] = frozendict(**{r"%Y": r"(?P<year>\d{4})",
                                                                    r"%m": r"(?P<month>[01]?\d)",
                                                                    r"%d": r"(?P<day>[0123]?\d)",
                                                                    r"%H": r"(?P<hour>[012]?\d)",
                                                                    r"%M": r"(?P<minute>[0-5]?\d)",
                                                                    r"%S": r"(?P<second>[0-5]?\d)",
                                                                    r"%f": r"(?P<microsecond>\d+)",
                                                                    r"%Z": r"(?P<tzinfo>[a-zA-Z]+([+-]\d{2}(\:\d{2})?)?)"})


def datetime_format_to_regex(in_format: str, flags: re.RegexFlag) -> re.Pattern:
    pattern_string = in_format
    for k, v in DATETIME_FORMAT_REGEX_MAPPING.items():
        pattern_string = pattern_string.replace(k, v)
    return re.compile(pattern_string, flags)

    # region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



from gidapptools.errors import MissingOptionalDependencyError


with MissingOptionalDependencyError.try_import("PySide6"):
    import PySide6


with MissingOptionalDependencyError.try_import("jinja2"):
    import jinja2



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from enum import Enum, auto, unique
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.utility.enums import BaseGidEnum

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


@unique
class MiscEnum(Enum):
    NOTHING = auto()
    ALL = auto()
    DEFAULT = auto()
    NOT_FOUND = auto()
    OPTIONAL = auto()
    ERROR = auto()
    AUTO = auto()

    def __repr__(self) -> str:
        return self.name

    def __rich__(self) -> str:
        return f"[u i medium_purple]{self.__repr__()}[/u i medium_purple]"

    def __bool__(self) -> bool:
        if self in {MiscEnum.NOTHING, MiscEnum.NOT_FOUND}:
            return False
        return True


class StringCase(BaseGidEnum):
    """
    _summary_

    - SNAKE: 'example_variable_name'
    - SCREAMING_SNAKE: 'EXAMPLE_VARIABLE_NAME'
    - CAMEL: 'exampleVariableName'
    - PASCAL: 'ExampleVariableName'
    - KEBAP: 'example-variable-name'
    - SPLIT: 'example variable name'
    - TITLE: 'Example Variable Name'
    - UPPER: 'EXAMPLE VARIABLE NAME'
    - BLOCK_UPPER: 'EXAMPLEVARIABLENAME'

    - CLASS: see `PASCAL`

    Args:
        BaseGidEnum (_type_): _description_
    """
    SNAKE = auto()
    SCREAMING_SNAKE = auto()
    CAMEL = auto()
    PASCAL = auto()
    KEBAP = auto()
    SPLIT = auto()
    TITLE = auto()
    UPPER = auto()
    BLOCK_UPPER = auto()
    # aliases
    CLASS = PASCAL


class FileTypus(BaseGidEnum):
    TXT = auto()
    MD = auto()
    INI = auto()
    JSON = auto()
    PY = auto()
    TOML = auto()
    YAML = auto()
    EXE = auto()
    BAT = auto()
    PS1 = auto()
    PNG = auto()
    JPEG = auto()
    GIF = auto()
    SVG = auto()

    # aliases
    CMD = BAT
    YML = YAML
    JPG = JPEG


# region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



from ._checks import check_has_data_set_methods, check_is_widget

from .standard_types import DefaultValueWidget, TextValueWidget, NumberValueWidget

from PySide6.QtWidgets import QWidget


DEFAULT_VALUE_WIDGET = DefaultValueWidget


WIDGET_MAP: dict[type:QWidget] = {}


def determine_value_widget(value: object) -> QWidget:

    return WIDGET_MAP.get(value.__class__, DEFAULT_VALUE_WIDGET)


for widget in (DefaultValueWidget, TextValueWidget, NumberValueWidget):
    widget.determine_value_widget_func = determine_value_widget
    assert check_is_widget(widget) is True
    assert check_has_data_set_methods(widget) is True
    for typus in widget.value_types:
        WIDGET_MAP[typus] = widget



"""
WiP.

Soon.
"""

# region [Imports]

import os
import re
import sys
import json
import queue
import math
import base64
import pickle
import random
import shelve
import dataclasses
import shutil
import asyncio
import logging
import sqlite3
import platform
from decimal import Decimal
import subprocess
import inspect

from time import sleep, process_time, process_time_ns, perf_counter, perf_counter_ns
from io import BytesIO, StringIO
from abc import ABC, ABCMeta, abstractmethod
from copy import copy, deepcopy
from enum import Enum, Flag, auto, unique
from pprint import pprint, pformat
from pathlib import Path
from string import Formatter, digits, printable, whitespace, punctuation, ascii_letters, ascii_lowercase, ascii_uppercase
from timeit import Timer
from typing import (TYPE_CHECKING, TypeVar, TypeGuard, TypeAlias, Final, TypedDict, Generic, Union, Optional, ForwardRef, final,
                    no_type_check, no_type_check_decorator, overload, get_type_hints, cast, Protocol, runtime_checkable, NoReturn, NewType, Literal, AnyStr, IO, BinaryIO, TextIO, Any)
from collections import Counter, ChainMap, deque, namedtuple, defaultdict
from collections.abc import (AsyncGenerator, AsyncIterable, AsyncIterator, Awaitable, ByteString, Callable, Collection, Container, Coroutine, Generator,
                             Hashable, ItemsView, Iterable, Iterator, KeysView, Mapping, MappingView, MutableMapping, MutableSequence, MutableSet, Reversible, Sequence, Set, Sized, ValuesView)
from zipfile import ZipFile, ZIP_LZMA
from datetime import datetime, timezone, timedelta
from tempfile import TemporaryDirectory
from textwrap import TextWrapper, fill, wrap, dedent, indent, shorten
from functools import wraps, partial, lru_cache, singledispatch, total_ordering, cached_property, cache
from contextlib import contextmanager, asynccontextmanager, nullcontext, closing, ExitStack, suppress
from statistics import mean, mode, stdev, median, variance, pvariance, harmonic_mean, median_grouped
from urllib.parse import urlparse
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, wait, as_completed, ALL_COMPLETED, FIRST_EXCEPTION, FIRST_COMPLETED

import PySide6
from PySide6 import (QtCore, QtGui, QtWidgets, Qt3DAnimation, Qt3DCore, Qt3DExtras, Qt3DInput, Qt3DLogic, Qt3DRender, QtAxContainer, QtBluetooth,
                     QtCharts, QtConcurrent, QtDataVisualization, QtDesigner, QtHelp, QtMultimedia, QtMultimediaWidgets, QtNetwork, QtNetworkAuth,
                     QtOpenGL, QtOpenGLWidgets, QtPositioning, QtPrintSupport, QtQml, QtQuick, QtQuickControls2, QtQuickWidgets, QtRemoteObjects,
                     QtScxml, QtSensors, QtSerialPort, QtSql, QtStateMachine, QtSvg, QtSvgWidgets, QtTest, QtUiTools, QtWebChannel, QtWebEngineCore,
                     QtWebEngineQuick, QtWebEngineWidgets, QtWebSockets, QtXml)

from PySide6.QtCore import (QByteArray, QCoreApplication, QDate, QDateTime, QEvent, QLocale, QMetaObject, QModelIndex, QModelRoleData, QMutex,
                            QMutexLocker, QObject, QPoint, QRect, QRecursiveMutex, QRunnable, QSettings, QSize, QThread, QThreadPool, QTime, QUrl,
                            QWaitCondition, Qt, QAbstractItemModel, QAbstractListModel, QAbstractTableModel, Signal, Slot)

from PySide6.QtGui import (QAction, QBrush, QClipboard, QMouseEvent, QTextBlock, QTextBlockFormat, QTextCharFormat, QSyntaxHighlighter, QTextOption, QTextCursor, QColor, QConicalGradient, QCursor, QFont, QFontDatabase, QFontMetrics, QGradient, QIcon, QImage,
                           QKeySequence, QLinearGradient, QPainter, QPalette, QPixmap, QRadialGradient, QTransform)

from PySide6.QtWidgets import (QApplication, QBoxLayout, QCheckBox, QColorDialog, QColumnView, QComboBox, QDateTimeEdit, QDialogButtonBox,
                               QDockWidget, QDoubleSpinBox, QFontComboBox, QFormLayout, QFrame, QGridLayout, QGroupBox, QHBoxLayout, QHeaderView,
                               QLCDNumber, QLabel, QLayout, QLineEdit, QListView, QListWidget, QMainWindow, QMenu, QMenuBar, QMessageBox,
                               QProgressBar, QProgressDialog, QPushButton, QSizePolicy, QSpacerItem, QSpinBox, QStackedLayout, QStackedWidget,
                               QStatusBar, QStyledItemDelegate, QSystemTrayIcon, QTabWidget, QTableView, QTextEdit, QTimeEdit, QToolBox, QTreeView,
                               QVBoxLayout, QWidget, QAbstractItemDelegate, QAbstractItemView, QAbstractScrollArea, QRadioButton, QFileDialog, QButtonGroup)

from gidapptools.gidapptools_qt.helper.syntax_highlighting.rules import ALL_RULES
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    from gidapptools.gidapptools_qt.abc.abstract_syntax_highlighting_rule import AbstractSyntaxHighlightRule

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class DebugValueHighlighter(QSyntaxHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.rules: dict[str:"AbstractSyntaxHighlightRule"] = {}

    def add_rule(self, rule: "AbstractSyntaxHighlightRule"):
        self.rules[rule.name] = rule

    def highlightBlock(self, text: str) -> None:
        for rule in self.rules.values():

            for _format in rule.apply(text, self):

                self.setFormat(*_format)


class DefaultValueWidget(QTextEdit):
    determine_value_widget_func: Callable[[object], QWidget] = None
    value_types: tuple[type] = tuple()

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.highlighter: QSyntaxHighlighter = None
        self.is_setup: bool = False

    def setup(self) -> Self:
        if self.is_setup is True:
            return self

        self.highlighter = DebugValueHighlighter(self)
        self.setup_highlighter()
        self.setFontFamily("Lucida Console")
        new_point_size = self.font().pointSize() * 1.25

        self.setFontPointSize(new_point_size)
        self.setReadOnly(True)
        self.setWordWrapMode(QTextOption.WrapMode.NoWrap)

        self.is_setup = True
        return self

    def set_data(self, data: object) -> None:
        self.setText(pformat(data))
        text_cursor = self.textCursor()
        block_format = QTextBlockFormat()

        block_format.setLineHeight(5, QTextBlockFormat.LineHeightTypes.LineDistanceHeight.value)

        text_cursor.clearSelection()
        text_cursor.select(QTextCursor.Document)
        text_cursor.mergeBlockFormat(block_format)
        self.setToolTip(str(data.__class__))

    def setData(self, data: object) -> None:
        self.set_data(data=data)

    def setup_highlighter(self):
        if self.highlighter is None:
            return
        for rule in ALL_RULES:
            self.highlighter.add_rule(rule())
        self.highlighter.setDocument(self.document())


class TextValueWidget(QTextEdit):
    determine_value_widget_func: Callable[[object], QWidget] = None
    value_types: tuple[type] = (str,)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.is_setup: bool = False

    def setup(self) -> Self:
        if self.is_setup is True:
            return self

        self.setFontFamily("Lucida Console")
        new_point_size = self.font().pointSize() * 1.25

        self.setFontPointSize(new_point_size)
        self.setReadOnly(True)
        self.setWordWrapMode(QTextOption.WrapMode.WrapAtWordBoundaryOrAnywhere)

        self.is_setup = True
        return self

    def set_data(self, data: str) -> None:
        self.setText(data)
        text_cursor = self.textCursor()
        block_format = QTextBlockFormat()

        block_format.setLineHeight(5, QTextBlockFormat.LineHeightTypes.LineDistanceHeight.value)

        text_cursor.clearSelection()
        text_cursor.select(QTextCursor.Document)
        text_cursor.mergeBlockFormat(block_format)
        self.setToolTip(str(data.__class__))

    def setData(self, data: object) -> None:
        self.set_data(data=data)


# class NumberValueWidget(QLineEdit):
#     determine_value_widget_func: Callable[[object], QWidget] = None
#     value_types: tuple[type] = (int, float)

#     def __init__(self, parent=None):
#         super().__init__(parent=parent)
#         self.is_setup: bool = False

#     def setup(self) -> Self:
#         if self.is_setup is True:
#             return self
#         self.setAlignment(Qt.AlignmentFlag.AlignCenter)
#         font = self.font()
#         font.setFamily("Lucida Console")
#         font.setPointSize(int(font.pointSize() * 1.25))
#         self.setFont(font)
#         self.setReadOnly(True)

#         self.is_setup = True
#         return self

#     def set_data(self, data: Union[int, float]) -> None:
#         self.setText(str(data))
#         self.setToolTip(str(data.__class__))

#     def setData(self, data: object) -> None:
#         self.set_data(data=data)


class NumberValueWidget(QLCDNumber):
    determine_value_widget_func: Callable[[object], QWidget] = None
    value_types: tuple[type] = (int, float, Decimal)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._raw_data: Union[int, float] = None
        self.is_setup: bool = False

    def setup(self) -> Self:
        if self.is_setup is True:
            return self
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        self.setFrameShape(QFrame.Shape.Box)
        self.setSegmentStyle(self.SegmentStyle.Flat)
        self.setFixedHeight(150)

        copy_action = QAction("copy", self)
        copy_action.triggered.connect(self.on_copy)
        self.addAction(copy_action)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.ActionsContextMenu)
        self.setSmallDecimalPoint(True)
        self.is_setup = True
        return self

    def set_data(self, data: Union[int, float]) -> None:
        self._raw_data = data
        digit_amount = len(str(self._raw_data).replace(".", ""))

        self.setDigitCount(digit_amount)
        if isinstance(data, Decimal):
            data = float(data)

        self.display(f"{data!s}")
        self.setToolTip(str(data.__class__))

    def setData(self, data: object) -> None:

        self.set_data(data=data)

    def on_copy(self, checked: bool):
        clipboard = QApplication.clipboard()
        value = str(self._raw_data)
        clipboard.setText(value)


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path
from threading import RLock

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import Qt, QCloseEvent
from PySide6.QtCore import Signal
from PySide6.QtWidgets import QWidget, QApplication

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class SecondaryWindow(QWidget):
    close_signal = Signal(QWidget)

    def __init__(self, parent: QWidget = None, f: Qt.WindowFlags = None, name: str = None) -> None:
        super().__init__(*[i for i in [parent, f] if i is not None])
        self.is_closed: bool = False
        if name is not None:
            self.setObjectName(name)
            self.setWindowTitle(name)

    @property
    def app(self) -> QApplication:
        return QApplication.instance()

    def closeEvent(self, event: QCloseEvent) -> None:
        super().closeEvent(event)
        self.close_signal.emit(self)
        self.is_closed = True

    def __repr__(self) -> str:

        return f'{self.__class__.__name__}'


class WindowReferenceKeeper:

    def __init__(self) -> None:
        self.lock = RLock()
        self._open_windows: list[SecondaryWindow] = []

    def show(self, window: SecondaryWindow) -> SecondaryWindow:
        if not isinstance(window, SecondaryWindow):
            raise TypeError(f"{self.__class__.__name__!r} can only handle windows that are a subclass of {SecondaryWindow.__name__!r}.")
        with self.lock:
            if window not in self._open_windows:
                self._open_windows.append(window)
            window.close_event.connect(self.on_window_close)
            window.show()
            return window

    def on_window_close(self, window: SecondaryWindow) -> None:
        with self.lock:
            try:
                self._open_windows.remove(window)
            except ValueError:
                pass

    def remove_already_closed(self) -> None:
        with self.lock:
            self._open_windows = [w for w in self._open_windows if w.is_closed is False]


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]
